/* 
	Some fixes by Juraj Michalek http://games.linux.sk
	NULL values removed by zero
*/

#include <time.h>
#include <string>
#include <fstream.h>
#include <allegro.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
using namespace std;

#define PARTICLENUM 1000
#define TOTAL_SPELLS 42
#define TOTAL_ITEMS 7
#define NUM_ENEMIES 6
#define LAST_LEVEL 19
#define FIRST_LEVEL 1  //for debugging purposes


char msg[20];
long gametime;  //for seeing how long each level lasted
int exitgame;
int gamestatus;
int initialized;
int score;
int level;
int died;
int temporary;
int temporary2;
int keydelay;
int MAXHEALTH;
int nextlevel;  //this is only used when returning from a warp level
int warplevel = 101;  //this gives the last warp level gone to
int warping;  //1 if warping forward, 0 if not, -1 if returning to the normal levels
int final_particle = 0;  //for keeping the particle loop as short as possible
int frames;
int fps;
int lframes;
int lfps;
int showfps;

volatile int counter;
volatile int fpscounter;

typedef struct effect     //particle types: 1) wavy vertical grey - until screen top
{                         //                2) scattered color
  int x;                  //                3) wavy vertical grey - life limit
  int y;                  //                4) sparkler color
  int xv;                 //                5) center-bound color
  int yv;                 //                6) still color
  int life;               //                7) moving color
  int exist;              //                8) flame trail
  int color;              //                9) ground flying up
  int type;               //               10) screen-filling ellipse
  int temp;
  int max;
} effect;

typedef struct mover
{
  int type;
  int health;
  int movecount;
  int attacked;
  int spellcount;
  int location;
  int oldlocation;
  int moving;
  int frame;
  int clip;
  int exist;
  int direction;
  int temp;
  int speed;
  int power;
  int trapped;
  int shielded;  //0 = unprotected, 1 = shield, 2 = barrier, 3 = both
  int shieldcount;  //counts down until shield/barrier runs out
} mover;

typedef struct list
{
  int type;
  int percent;
  int active;
  int speed;
  int time;
  int element;
  int power;
  int kind;
  char name[15];
  string description[5];
  int need[TOTAL_SPELLS];  //stores the required spells before available for learning
  int buy[4];  //the amount of orbs required to learn the spell
} list;

typedef struct grid
{
  int status;
  int object;
  int spell;
  int frame;
  int x;
  int y;
  int up;
  int down;
  int left;
  int right;
} grid;

typedef struct rspell  //ranged spells
{
  int type;
  int location;
  int oldlocation;
  int moving;
  int frame;
  int clip;
  int exist;
  int direction;
  int speed;
  int power;
  int movecount;
  int element;
} rspell;

typedef struct awardsgiven  //for the bonus screen
{
  string description;
  int points;
  int oldpoints;
  int old;
} awardsgiven;

typedef struct description
{
  string disc[4];
  int strong[4];
  int weak[4];
} description;

effect particle[PARTICLENUM];
effect graphic[81];
mover player[40];  //player[0] is the actual player.  The rest are summoned allies.
mover enemy[15];
list spell[6];
list archive[TOTAL_SPELLS];
grid tile[81];
rspell ranged[10];
awardsgiven award[5];
description enemydisc[NUM_ENEMIES];

int typespeed[15];
int typehealth[15];
int typepower[15];
int oldscores[10];
int orb[4];
int warp;
string names[10];

void titlescreen();
void menu();
void logic();
void init_particle(int type, int x, int y, int color, int xv, int yv);
void init_particle(int type, int x, int y, int color);
void doparticles();
void show_startmenu();
void options();
void show_highscores();
void end_of_level();
void levelinit();
void draw_board();
void draw_all();
void draw_object(int tile, int type);
void draw_mover(int tile, int type, int offset_x, int offset_y, int direction, int anim);
void draw_stats();
void draw_spells();
void check_keys();
void reset_all();
void reset_grid();
void do_ai();
void create_enemy(int type, int tile);
void initialize_spells();
void draw_ranged_spell(rspell current, int offset_x, int offset_y);
void check_for_win();
void reset_table();
void post_score(int loc, string entry);
void enter_highscore();
void query_exit();
void load_game();
void kill_particles();
void show_chosenspell(int spl);
void buy_spell(int spl);
void drop_item(int loc, int type);
void grab_item(int type);
void add_warp();
void draw_trap(mover current);
void draw_shield(int shieldtype, int x, int y);
void do_description();
void set_trap(int tilenum, int type);
void draw_sample_enemy(int type, int x, int y);
int set_brightness();
int find_free();
int find_free_ranged();
int find_free_graphic();
int target_player(mover);
int check_move(int tile, int direction);
int invert_move(int direction);
int rotate_move(int direction);
int check_for_loss();
int check_highscores();
int numpoints(int type);
int increment_chosenspell(int spl, int inc, int buy);
int find_first_learner();
int check_price(int spl);
int check_warp(int func);  //func changes what check_warp returns
int get_tile(int tilenum, int direction);
int enemy_on_tile(int tilenum);
int player_on_tile(int tilenum);
string element(int type);
awardsgiven get_awards(awardsgiven a, int num);
mover calc_mover(mover current);
mover calc_mover_ai(mover current);
mover cast_spell(mover current, int spelltype);
mover deal_damage(mover current, int pow, int element);
mover cast_enemy_spell(mover current, int type);

BITMAP *back_buffer;
DATAFILE *fonts;
DATAFILE *graphics;
DATAFILE *palette;
DATAFILE *movers;
DATAFILE *spells;
DATAFILE *tiles;
DATAFILE *bigspark;
DATAFILE *statbar;
DATAFILE *mage;
DATAFILE *slime;
DATAFILE *fire;
DATAFILE *magewalk;
DATAFILE *magewalk1;
DATAFILE *magewalk2;
DATAFILE *magewalk3;
DATAFILE *magewalk4;
DATAFILE *goblin;
DATAFILE *gobwalk1;
DATAFILE *gobwalk2;
DATAFILE *gobwalk3;
DATAFILE *gobwalk4;
DATAFILE *menupic;
DATAFILE *orbs;
DATAFILE *jab;
DATAFILE *misc;
DATAFILE *spike;
DATAFILE *needle;
DATAFILE *lightning;
DATAFILE *firepit;
DATAFILE *flamejet;
DATAFILE *ice;
DATAFILE *shield;
DATAFILE *ball_lit;
DATAFILE *barrier;
DATAFILE *power;
DATAFILE *quake;
DATAFILE *encase;
DATAFILE *amber;
DATAFILE *fireball;
DATAFILE *mole;
DATAFILE *molepop;
DATAFILE *clay;
DATAFILE *claydl;
DATAFILE *clayur;
DATAFILE *phantom;
DATAFILE *imp;
DATAFILE *imp1;
DATAFILE *imp2;
DATAFILE *imp3;
DATAFILE *imp4;

void timer_handler()
{
  counter++;
}
END_OF_FUNCTION(timer_handler);

void fps_handler()
{
  fpscounter++;
}
END_OF_FUNCTION(fps_handler);

int main()
{
  BEGIN_COLOR_DEPTH_LIST
    COLOR_DEPTH_8
  END_COLOR_DEPTH_LIST
  LOCK_VARIABLE(counter);
  LOCK_FUNCTION(timer_handler);


  srandom(time(NULL));

  allegro_init();
  install_keyboard();
  install_timer();
  install_int_ex(timer_handler, BPS_TO_TIMER(100));
  install_int_ex(fps_handler, BPS_TO_TIMER(1));
  if (set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0) != 0)
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
    return 1;
  }
  textout_centre(screen, font, "Loading...", 320, 200, 15);
  fonts = load_datafile("fonts.dat");
  if (!fonts)
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Unable to find 'fonts.dat'\n%s\n", allegro_error);
    return 1;
  }
  ifstream tilegrid("tilegrid.dat");
  if (tilegrid.bad())
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Unable to find 'tilegrid.dat'\n%s\n", allegro_error);
    return 1;
  }
  temporary = 0;
  while (temporary < 81)
  {
    tilegrid >> tile[temporary].up;
    tilegrid >> tile[temporary].right;
    tilegrid >> tile[temporary].down;
    tilegrid >> tile[temporary].left;
    temporary++;
  }
  ifstream enemies("enemy.dat");
  if (enemies.bad())
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Unable to find 'enemy.dat'\n%s\n", allegro_error);
    return 1;
  }
  string trash;  //junk string
  temporary = 0;
  while (temporary < NUM_ENEMIES)
  {
    enemies >> trash;  //skips the enemy name in the file
    enemies >> typespeed[temporary];
    enemies >> typehealth[temporary];
    enemies >> typepower[temporary];
    getline(enemies, trash);  //I'm not clear on why this is necessary.  It just is.
    temporary2 = 0;
    while (temporary2 < 4)
    {
      getline(enemies, enemydisc[temporary].disc[temporary2]);
      temporary2++;
    }
    temporary2 = 0;
    while (temporary2 < 4)
    {
      enemies >> enemydisc[temporary].strong[temporary2];
      temporary2++;
    }
    temporary2 = 0;
    while (temporary2 < 4)
    {
      enemies >> enemydisc[temporary].weak[temporary2];
      temporary2++;
    }
    temporary++;
  }
  ifstream oldtable("hiscores.dat");
  if (oldtable.bad())
  {
    reset_table();
  }
  temporary = 0;
  while (temporary < 10)
  {
    oldtable >> names[temporary];
    oldtable >> oldscores[temporary];
    temporary++;
  }
  ifstream archives("spells.dat");
  if (archives.bad())
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Unable to find 'spells.dat'\n%s\n", allegro_error);
    return 1;
  }
  temporary = 0;
  while (temporary < TOTAL_SPELLS)
  {
    archives >> archive[temporary].name;
    archives >> archive[temporary].type;
    archives >> archive[temporary].speed;
    archives >> archive[temporary].element;
    archives >> archive[temporary].power;
    archives >> archive[temporary].time;
    archives >> archive[temporary].kind;
    getline(archives, trash);  //I'm not clear on why this is necessary.  It just is.
    temporary2 = 0;
    while (temporary2 < 5)
    {
      getline(archives, archive[temporary].description[temporary2]);
      temporary2++;
    }
    temporary2 = 0;
    while (temporary2 < TOTAL_SPELLS)
    {
      archives >> archive[temporary].need[temporary2];
      if (archive[temporary].need[temporary2] == -1)
      {
        temporary2 = TOTAL_SPELLS;  //exit the loop
      }
      else
      {
        temporary2++;
      }
    }
    temporary2 = 0;
    while (temporary2 < 4)
    {
      archives >> archive[temporary].buy[temporary2];
      temporary2++;
    }
    temporary++;
  }
  graphics = load_datafile("graphics.dat");
  if (!graphics)
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Unable to find 'graphics.dat'\n%s\n", allegro_error);
    return 1;
  }
  movers = (DATAFILE *)graphics[0].dat;
  spells = (DATAFILE *)graphics[1].dat;
  menupic = (DATAFILE *)graphics[2].dat;
  orbs = (DATAFILE *)graphics[3].dat;
  misc = (DATAFILE *)graphics[4].dat;
  bigspark = (DATAFILE *)movers[0].dat;
  statbar = (DATAFILE *)movers[1].dat;
  tiles = (DATAFILE *)movers[2].dat;
  mage = (DATAFILE *)movers[3].dat;
  slime = (DATAFILE *)movers[4].dat;
  magewalk = (DATAFILE *)movers[5].dat;
  goblin = (DATAFILE *)movers[6].dat;
  mole = (DATAFILE *)movers[7].dat;
  clay = (DATAFILE *)movers[8].dat;
  phantom = (DATAFILE *)movers[9].dat;
  imp = (DATAFILE *)movers[10].dat;
  gobwalk1 = (DATAFILE *)goblin[0].dat;
  gobwalk2 = (DATAFILE *)goblin[1].dat;
  gobwalk3 = (DATAFILE *)goblin[2].dat;
  gobwalk4 = (DATAFILE *)goblin[3].dat;
  magewalk1 = (DATAFILE *)magewalk[0].dat;
  magewalk2 = (DATAFILE *)magewalk[1].dat;
  magewalk3 = (DATAFILE *)magewalk[2].dat;
  magewalk4 = (DATAFILE *)magewalk[3].dat;
  molepop = (DATAFILE *)mole[1].dat;
  claydl = (DATAFILE *)clay[1].dat;
  clayur = (DATAFILE *)clay[2].dat;
  imp1 = (DATAFILE *)imp[0].dat;
  imp2 = (DATAFILE *)imp[1].dat;
  imp3 = (DATAFILE *)imp[2].dat;
  imp4 = (DATAFILE *)imp[3].dat;
  fire = (DATAFILE *)spells[0].dat;
  jab = (DATAFILE *)spells[1].dat;
  spike = (DATAFILE *)spells[2].dat;
  needle = (DATAFILE *)spells[3].dat;
  lightning = (DATAFILE *)spells[4].dat;
  firepit = (DATAFILE *)spells[5].dat;
  flamejet = (DATAFILE *)spells[6].dat;
  ice = (DATAFILE *)spells[7].dat;
  shield = (DATAFILE *)spells[8].dat;
  ball_lit = (DATAFILE *)spells[9].dat;
  barrier = (DATAFILE *)spells[10].dat;
  power = (DATAFILE *)spells[11].dat;
  quake = (DATAFILE *)spells[12].dat;
  encase = (DATAFILE *)spells[13].dat;
  amber = (DATAFILE *)spells[14].dat;
  fireball = (DATAFILE *)spells[15].dat;
  palette = load_datafile("palette.dat");
  if (!palette)
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("Unable to find 'palette.dat'\n%s\n", allegro_error);
    return 1;
  }
  set_palette((RGB *)palette[0].dat);
  back_buffer = create_bitmap(640, 480);
  clear(back_buffer);
  color_map = (COLOR_MAP *)palette[1].dat;
  exitgame = 0;
  gamestatus = 0;
  counter = 0;
  frames = 0;
  fps = 0;
  showfps = 0;
  while (exitgame == 0)
  {
    if (key[KEY_B]) //loop breaker button
    {
      exitgame = 1;
    }
    if (key[KEY_K])
    {
      player[0].health = 0; //kill button
    }
    if (key[KEY_S])  //grab a screenshot
    {
      save_bitmap("scrshot.bmp", back_buffer, (RGB *)palette[0].dat);
    }
    if (key[KEY_T])  //toggle the framerate
    {
      if (showfps == 1)
      {
        showfps = 0;
      }
      else
      {
        showfps = 1;
      }
    }
    frames++;
    while (counter > 1)
    {
      switch (gamestatus)
      {
        case 0:
          titlescreen();
          break;
        case 1:
          menu();
          break;
        case 2:
          logic();
          break;
        case 3:
          options();
          break;
        case 4:
          show_highscores();
          break;
        case 5:
          enter_highscore();
          break;
        case 6:
          load_game();
          break;
        case 7:
          end_of_level();
          break;
      }
      lframes++;
      counter -= 2;
      if (fpscounter >= 1)
      {
        fpscounter = 0;
        fps = frames;
        lfps = lframes;
        lframes = 0;
        frames = 0;
      }
    }
    if (showfps == 1)
    {
      sprintf(msg, "Logic FPS:  %-1d", lfps);
      textout(back_buffer, font, msg, 5, 5, 15);
      sprintf(msg, "Screen FPS: %-1d", fps);
      textout(back_buffer, font, msg, 5, 15, 15);
    }
    blit(back_buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
  }
  textout_centre(screen, font, "Goodbye", 320, 240, 15);
  ofstream newtable("hiscores.dat");
  temporary = 0;
  while (temporary < 10)
  {
    newtable << names[temporary];
    newtable << "\n";
    newtable << oldscores[temporary];
    newtable << "\n";
    temporary++;
  }
  newtable.close();
  destroy_bitmap(back_buffer);
  unload_datafile(fonts);
  unload_datafile(graphics);
  unload_datafile(palette);
  remove_keyboard();
  remove_int(timer_handler);
  remove_int(fps_handler);
  remove_timer();
  remove_sound();
  allegro_exit();
  return 0;
}


END_OF_MAIN();


void titlescreen()
{
  static int count = 0;
  if (count == 0)
  {
    kill_particles(); //wipe out any active particles that might linger on otherwise
  }
  int z = 0;
  clear(back_buffer);
  count++;
  if (count > 20)
  {
    while (z < 30)
    {
      init_particle(1, (random()%175) + 235, (random()%30) + 30, 0);
      z++;
    }
    if (count < 60)
    {
      textout_centre(back_buffer, (FONT *)fonts[0].dat, "Sole Mage", 320, 80 - count, 15);
      rectfill(back_buffer, 230, 55, 410, 110, 0);
      while (z < 50)
      {
        init_particle(1, (random()%175) + 235, 55, 0);
        z++;
      }
    }
    else
    {
      textout_centre(back_buffer, (FONT *)fonts[0].dat, "Sole Mage", 320, 20, 15);
      if (count > 120)
      {
        show_startmenu();
      }
    }
  }
  if (key[KEY_R]) //redo effect
  {
    kill_particles();
    count = 0;
  }
  if ((key[KEY_ESC]) && (count > 30))  //gives a small delay so quitting the high
  {                                    //scores table doesn't make you exit the game
    exitgame = 1;
  }
  if (gamestatus != 0)
  {
    count = 0;
    kill_particles();
  }
  doparticles();
}





void show_startmenu()
{
  static int choice = 0;
  static int delay = 15;
  static int frame = 0;
  int clip;
  int i;
  textout_centre(back_buffer, (FONT *)fonts[1].dat, "Start Game", 320, 250, 10);
  textout_centre(back_buffer, (FONT *)fonts[1].dat, "Load Game", 320, 270, 10);
  textout_centre(back_buffer, (FONT *)fonts[1].dat, "Options", 320, 290, 10);
  textout_centre(back_buffer, (FONT *)fonts[1].dat, "High Scores", 320, 310, 10);
  if (delay <= 0)
  {
    if (key[KEY_DOWN])
    {
      delay = 5;
      if (choice < 3)
      {
        choice++;
      }
    }
    if (key[KEY_UP])
    {
      delay = 5;
      if (choice > 0)
      {
        choice--;
      }
    }
  }
  else
  {
    delay--;
  }
  if (key[KEY_ENTER])
  {
    switch(choice)
    {
      case 0:
        gamestatus = 1;
        initialized = 0;
        reset_all();
        keydelay = 10;        
        break;
      case 1:
        gamestatus = 6;
        initialized = 0;
        reset_all();
        break;
      case 2:
        gamestatus = 3;
        break;
      case 3:
        gamestatus = 4;
        break;
    }
    choice = 0;
  }
  frame++;
  if (frame > 30)
  {
    frame = 0;
  }
  clip = frame / 3;
  draw_sprite(back_buffer, (BITMAP *)bigspark[clip].dat, 130, 205 + (20 * choice));
}





void reset_all()
{
  int q = 0;
  keydelay = 0;
  score = 0;
  died = 0;
  level = FIRST_LEVEL;
  spell[0].active = 1;
  spell[0].type = 1;
  spell[1].active = 1;
  spell[1].type = 2;
  while (q < TOTAL_SPELLS)
  {
    if (q < 2)
    {
      archive[q].active = 1;
    }
    else
    {
      archive[q].active = 0;
    }
    q++;
  }
  q = 0;
  while (q < 4)
  {
    orb[q] = 50;
    q++;
  }
  q = 0;
  while (q < 5)
  {
    award[q].old = -1;
    award[q].oldpoints = 0;
    award[q].description = "-";
    q++;
  }
  warp = 0;
  MAXHEALTH = 10;
  reset_grid();
}





void reset_grid()
{
  int i = 0;
  player[0].exist = 1;
  player[0].location = 64;
  player[0].oldlocation = 64;
  player[0].type = 0;
  player[0].moving = -1;
  player[0].movecount = 0;
  player[0].direction = 1;
  player[0].health = MAXHEALTH;
  player[0].shielded = 0;
  player[0].shieldcount = 0;
  player[0].trapped = 0;
  gametime = 0;
  while (i < 81)
  {
    tile[i].object = -1;
    tile[i].status = -1;
    tile[i].frame = 0;
    tile[i].spell = -1;
    graphic[i].exist = 0;
    graphic[i].temp = 0;
    graphic[i].max = 0;
    graphic[i].life = 0;
    if (i < 40)
    {
      if (i != 0)
      {
        player[i].exist = 0;
      }
      if (i < 15)
      {
        enemy[i].exist = 0;
        enemy[i].moving = -1;
        enemy[i].movecount = 0;
        enemy[i].trapped = 0;
        enemy[i].movecount = 0;
        enemy[i].attacked = 0;
        enemy[i].frame = 0;
        enemy[i].shielded = 0;
        enemy[i].shieldcount = 0;
        if (i < 10)
        {
          ranged[i].exist = 0;
        }
      }
    }
    i++;
  }
  i = 0;
  while (i < PARTICLENUM)
  {
    particle[i].exist = 0;
    i++;
  }
}





void options()
{
  clear(back_buffer);
  static int choice = 2;
  static int delay = 15;
  static int frame = 0;
  static int shown = -1;
  if (shown == -1)
  {
    int clip;
    int i = 0;
    while (i < PARTICLENUM)
    {
      particle[i].exist = 0;
      i++;
    }
    textout(back_buffer, (FONT *)fonts[1].dat, "Set Brightness Levels", 270, 250, 10);
    textout(back_buffer, (FONT *)fonts[1].dat, "Enter Cheat", 270, 270, 10);
    textout(back_buffer, (FONT *)fonts[1].dat, "Back", 270, 290, 10);
    if (delay <= 0)
    {
      if (key[KEY_DOWN])
      {
        delay = 5;
        if (choice < 2)
        {
          choice++;
        }
      }
      if (key[KEY_UP])
      {
        delay = 5;
        if (choice > 0)
        {
          choice--;
        }
      }
      if (key[KEY_ESC])
      {
        delay = 5;
        choice = 2;
      }
      if (key[KEY_ENTER])
      {
        switch(choice)
        {
          case 0:
            shown = 1;
            break;
          case 1:
            break;
          case 2:
            gamestatus = 0;
            delay = 15;
            break;
        }
      }
    }
    else
    {
      delay--;
    }
    frame++;
    if (frame > 30)
    {
      frame = 0;
    }
    clip = frame / 3;
    draw_sprite(back_buffer, (BITMAP *)bigspark[clip].dat, 120, 205 + (20 * choice));
  }
  else if (shown == 1)
  {
    shown = set_brightness();
    if (shown == -1)
    {
      delay = 15;
    }
  }
}





int set_brightness()
{
  rect(back_buffer, 110, 15, 530, 65, 2);
  textout_centre(back_buffer, (FONT *)fonts[0].dat, "Can you see this?", 320, 20, 21);
  textout(back_buffer, font, "Turn your brightness settings down until the", 110, 150, 2);
  textout(back_buffer, font, "message at the top of your screen just disappears.", 110, 160, 2);
  textout(back_buffer, font, "Press ESC to return to the options menu.", 110, 200, 2);
  if (key[KEY_ESC])
  {
    return -1;
  }
  else
  {
    return 1;
  }
}





void menu()
{
  initialize_spells();  //this is so the spell names appear the first time and...
  int o;                //...so that the names change when you select different spells
  int tnum;
  clear_to_color(back_buffer, 128);
  static int choice = 0;
  static int active = 0;
  static int chosenspell = 0;
  static int foundfirst = 0;
  static int boughtone = 0;
  static int foresaw = 0;  //records whether foresight has been used once this time
  rect(back_buffer, 40, 40, 230, 115, 43);
  rectfill(back_buffer, 41, 41, 229, 114, 0);
  rect(back_buffer, 39, 14, 245, 35, 43);
  rectfill(back_buffer, 40, 15, 244, 34, 0);
  sprintf(msg, "Next Level: %-1d", level);
  textout(back_buffer, (FONT *)fonts[2].dat, msg, 40, 15, 15);
  rect(back_buffer, 399, 14, 635, 35, 43);
  rectfill(back_buffer, 400, 15, 634, 34, 0);
  sprintf(msg, "Score: %-1d", score);
  textout(back_buffer, (FONT *)fonts[2].dat, msg, 400, 15, 15);
  textout(back_buffer, font, "Play next level", 70, 50, 15);
  textout(back_buffer, font, "Choose Spells", 70, 65, 15);
  textout(back_buffer, font, "Learn New Spell", 70, 80, 15);
  textout(back_buffer, font, "Save Your Progress", 70, 95, 15);
  draw_sprite(back_buffer, (BITMAP *)statbar[1].dat, 52, 52 + (15 * choice));
  o = 0;
  while (o < 5)
  {
    draw_sprite(back_buffer, (BITMAP *)statbar[0].dat, 480, 320 + (30 * o));
    if (spell[o].active == 1)
    {
      rectfill(back_buffer, 484, 324 + (30 * o), 625, 345 + (30 * o), 0);
      textout_centre(back_buffer, font, spell[o].name, 555, 331 + (30 * o), 15);
    }
    o++;
  }
  o = 0;
  rect(back_buffer, 10, 455, 220, 478, 43);
  rectfill(back_buffer, 11, 456, 219, 477, 0);
  while (o < 4)
  {
    draw_sprite(back_buffer, (BITMAP *)orbs[o].dat, 20 + (50 * o), 460);
    sprintf(msg, "%-1d", orb[o]);
    textout(back_buffer, font, msg, 40 + (50 * o), 465, 15);
    o++;
  }
  if (archive[19].active == 1)  //if 'scan' has been learned
  {
    rect(back_buffer, 10, 400, 210, 423, 43);
    rectfill(back_buffer, 11, 401, 209, 422, 0);
    textout(back_buffer, font, "Press 'E' for enemy data", 15, 408, 15);
    if (key[KEY_E])
    {
      active = 4;
      keydelay = 10;
    }
  }
  if (archive[23].active == 1)  //if 'foresight' has been learned
  {
    rect(back_buffer, 10, 428, 225, 450, 43);
    rectfill(back_buffer, 11, 429, 224, 449, 0);
    textout(back_buffer, font, "Press 'F' for future sight", 15, 435, 15);
    if (key[KEY_F])
    {
      active = 5;
      keydelay = 10;
    }
  }
  if (boughtone > 1)
  {
    rect(back_buffer, 235, 230, 350, 250, 43);
    rectfill(back_buffer, 236, 231, 349, 249, 0);
    textout(back_buffer, font, "Spell Learned", 240, 235, 15);
    boughtone--;
  }
  switch(active)
  {
    case 0:
      init_particle(2, 55, 54 + (15 * choice), 15);
      init_particle(2, 55, 54 + (15 * choice), 15);
      if (keydelay == 0)
      {
        if (key[KEY_UP])
        {
          if (choice > 0)
          {
            choice--;
            keydelay = 10;
          }
        }
        if (key[KEY_DOWN])
        {
          if (choice < 3)
          {
            choice++;
            keydelay = 10;
          }
        }
        if (key[KEY_ENTER])
        {
          switch(choice)
          {
            case 0:
              initialized = 0;
              initialize_spells();
              gamestatus = 2;
              keydelay = 0;
              foresaw = 0;  //to redo foresight next level menu
              reset_grid();
              fade_out(8);
              counter = 0;
              break;
            case 1:
              active = 1;
              keydelay = 10;
              chosenspell = 0;
              break;
            case 2:
              active = 2;
              keydelay = 10;
              break;
            case 3:
              active = 3;
              keydelay = 10;
              break;
          }
        }
        if (key[KEY_ESC])
        {
          query_exit();
          if (gamestatus == 0)  //if the player has chosen to exit
          {                     //reset the default choice
            choice = 0;
            kill_particles();   //query_exit already calls kill_particles
          }                     //but I just wanted to make sure
        }
      }
      else
      {
        keydelay--;
      }
      break;
    case 1:  //choose spell
      rect(back_buffer, 235, 40, 620, 200, 43);
      rectfill(back_buffer, 236, 41, 619, 199, 0);
      show_chosenspell(chosenspell);
      if (keydelay == 0)
      {
        if ((key[KEY_ESC]) || (key[KEY_ENTER]))
        {
          active = 0;
          keydelay = 10;
          chosenspell = 0;
        }
        if (key[KEY_RIGHT])
        {
          chosenspell = increment_chosenspell(chosenspell, 1, 1);
          keydelay = 10;
        }
        if (key[KEY_LEFT])
        {
          chosenspell = increment_chosenspell(chosenspell, -1, 1);
          keydelay = 10;
        }
        if (key[KEY_1])
        {
          spell[0].type = chosenspell + 1;  //the "+ 1" is there because spell
          keydelay = 10;                    //types start at 1 while chosenspell
        }                                   //starts at 0
        if (key[KEY_2])
        {
          spell[1].type = chosenspell + 1;
          keydelay = 10;
        }
        if (key[KEY_3])
        {
          spell[2].type = chosenspell + 1;
          keydelay = 10;
        }
        if (key[KEY_4])
        {
          spell[3].type = chosenspell + 1;
          keydelay = 10;
        }
        if (key[KEY_5])
        {
          spell[4].type = chosenspell + 1;
          keydelay = 10;
        }
      }
      else
      {
        keydelay--;
      }
      break;
    case 2:  //learn spell
      rect(back_buffer, 235, 40, 620, 220, 43);
      rectfill(back_buffer, 236, 41, 619, 219, 0);
      if (foundfirst == 0)
      {
        chosenspell = find_first_learner();
        foundfirst = 1;
      }
      if (chosenspell != -1)
      {
        show_chosenspell(chosenspell);
      }      
      if (keydelay == 0)
      {
        if (key[KEY_ESC])
        {
          active = 0;
          keydelay = 10;
          foundfirst = 0;
        }
        if (key[KEY_LEFT])
        {
          chosenspell = increment_chosenspell(chosenspell, 1, 0);
          keydelay = 10;
        }
        if (key[KEY_RIGHT])
        {
          chosenspell = increment_chosenspell(chosenspell, -1, 0);
          keydelay = 10;
        }
        if (key[KEY_ENTER])
        {
          if (check_price(chosenspell) == 1)
          {
            buy_spell(chosenspell);
            boughtone = 40;
          }
        }
      }
      else
      {
        keydelay--;
      }
      break;
    case 3:  //save game
      if (keydelay == 0)
      {
        static int chosen = -1;
        static string savedname[10];
        static int save_choice = 0;
        int i;
        if (chosen == -1)
        {
          i = 0;
          while (i < 10)
          {
            sprintf(msg, "player%-1d.sav", i);
            if (exists(msg))
            {
              ifstream savedgame(msg);        //if the file exists, load its name
              savedgame >> savedname[i];      //property onto the savedname array
            }
            else
            {
              savedname[i] = "-EMPTY-";       //if it doesn't exist, make it "-EMPTY-"
            }
            i++;
          }
          chosen = 0;
        }
        else if (chosen == 0)
        {
          i = 0;
          static int nkeydelay = 0;
          int loc_x = 50;            //all this stuff is for choosing the save file
          int loc_y = 150;           //to save to, and for inputting the name property
          int lx = 50;
          int ly = 150;
          rect(back_buffer, 45, 140, 255, 225, 15);
          rectfill(back_buffer, 51, 141, 254, 224, 0);
          ly = 145 + (15 * save_choice);                     //this
          if (save_choice > 4)                               //code
          {                                                  //draws
            lx = 150;                                        //the
            ly -= 75;                                        //file
          }                                                  //selection
          rect(back_buffer, lx, ly, lx + 100, ly + 15, 14);  //rectangle
          while (i < 10)
          {
            textout_centre(back_buffer, font, savedname[i].c_str(), loc_x + 50, loc_y, 15);
            loc_y += 15;
            if (loc_y >= 225)
            {
              loc_y = 150;
              loc_x = 150;
            }
            i++;
          }
          if (nkeydelay == 0)
          {
            if ((key[KEY_UP]) && (save_choice > 0))
            {
              save_choice--;
              nkeydelay = 10;
            }
            if ((key[KEY_DOWN]) && (save_choice < 9))
            {
              save_choice++;
              nkeydelay = 10;
            }
            if ((key[KEY_RIGHT]) && (save_choice < 5))
            {
              save_choice += 5;
              nkeydelay = 10;
            }
            if ((key[KEY_LEFT]) && (save_choice > 4))
            {
              save_choice -= 5;
              nkeydelay = 10;
            }
            if (key[KEY_ENTER])
            {
              chosen = 1;
            }
            if (key[KEY_ESC])
            {
              chosen = -1;
              active = 0;
              save_choice = 0;
              keydelay = 10;
            }
          }
          else
          {
            nkeydelay--;
          }
        }
        else
        {
          rect(screen, 10, 250, 140, 264, 15);
          rectfill(screen, 11, 251, 139, 263, 0);
          textout(screen, font, "Insert name:", 12, 252, 15);
          //-------------------------
          int x = 111;  //This code was blatantly ripped from a post on the Allegro.cc forums.
          int y = 250;  //It was posted by Chris Barry (23yrold3yrold), and then tweaked for
          int w = 140;  //my use (by me, of course).
          int h = 15;
          int leng = 10;
          i = 0;
          int k;
          int cancel = 0;
          int ok = 0;
          char name[12];
          BITMAP *bmp;
          bmp = create_bitmap(w, h);
          clear(bmp);
          rect(bmp, 0, 0, w - 1, h - 1, 2);
          name[0] = '\0';
          rectfill(bmp, 1, 1, w - 2, h - 2, makecol(50, 50, 50));
          textprintf(bmp, font, 2, 2, 15, "%s", name);//makecol(0, 200, 0)
          draw_sprite(screen, bmp, x, y);
          clear_keybuf();
          while (ok == 0)
          {
            k = readkey();
            if (((k >> 8) == KEY_ENTER))
            {
              if (i > 0)
              {
                ok = 1;
              }
            }
            else if (((k >> 8) == KEY_BACKSPACE) && i > 0)
            {
              i--;
              name[i] = '\0';
              }
            else if (((k >> 8) == KEY_BACKSPACE) && i <= 0)
            {
              //do nothing
            }
            else if (((k >> 8) == KEY_ESC))
            {
              ok = 1;
              cancel = 1;
              save_choice = 0;
            }
            else if (((k >> 8) == KEY_TAB) || ((k >> 8) == KEY_SPACE))
            {
              //do nothing
            }
            else
            {
              if (i < leng)
              {
                name[i] = k & 0xFF;
                name[i + 1] = '\0';
                i++;
              }
            }
            rectfill(bmp, 1, 1, w - 2, h - 2, makecol(50, 50, 50));
            textprintf(bmp, font, 2, 2, makecol(0, 200, 0), "%s", name);
            draw_sprite(screen, bmp, x, y);
          }
          counter = 0;
          //-------------------
          active = 0;
          keydelay = 15;
          chosen = -1;
          if (cancel != 1)
          {
            sprintf(msg, "player%-1d.sav", save_choice);
            int s = 0;
            ofstream savegame(msg);
            savegame << name << "\n";
            while (s < TOTAL_SPELLS)
            {
              savegame << archive[s].active << "\n";
              s++;
            }
            savegame << level << "\n";
            s = 0;
            while (s < 4)
            {
              savegame << orb[s] << "\n";
              s++;
            }
            savegame << player[0].health << "\n";
            savegame << warp << "\n";
            s = 0;
            while (s < 5)
            {
              savegame << spell[s].type << "\n";
              s++;
            }
            savegame << score << "\n";
            savegame << MAXHEALTH << "\n";
            savegame << nextlevel << "\n";
            savegame << warplevel << "\n";
            savegame << warping << "\n";
            s = 0;
            while (s < 5)
            {
              savegame << award[s].oldpoints << "\n";
              savegame << award[s].old << "\n";
              s++;
            }
          }
        }
      }
      else
      {
        keydelay--;
      }
      break;
    case 4:  //enemy scan
      if (key[KEY_ESC])
      {
        active = 0;
        keydelay = 10;
      }
      do_description();
      break;
    case 5:  //foresight
      static int seelev = level;
      if (keydelay == 0)
      {
        if (key[KEY_ESC])
        {
          active = 0;
          keydelay = 10;
        }
        if (archive[38].active == 1)  //if clairvoyance has been learned
        {
          if ((key[KEY_LEFT]) && (seelev > 1))
          {
            seelev--;
            keydelay = 10;
            foresaw = 0;
            reset_grid();
          }
          if ((key[KEY_RIGHT]) && (seelev < LAST_LEVEL - 1) && (seelev < level + 4))
          {
            seelev++;
            keydelay = 10;
            foresaw = 0;
            reset_grid();
          }
        }
      }
      else
      {
        keydelay--;
      }
      if ((foresaw == 0) && (seelev < LAST_LEVEL))  //second part keeps foresight from
      {                                             //going to the high score table
        o = level;
        level = seelev;
        levelinit();
        level = o;
        foresaw = 1;
      }
      draw_board();
      if (archive[38].active == 1)
      {
        rect(back_buffer, tile[74].x + 5, tile[80].y + 5, tile[74].x + 70, tile[80].y + 25, 43);
        rectfill(back_buffer, tile[74].x + 6, tile[80].y + 6, tile[74].x + 69, tile[80].y + 24, 0);
        sprintf(msg, "%-1d", seelev);
        textout_centre(back_buffer, (FONT *)fonts[1].dat, msg, tile[74].x + 38, tile[80].y + 8, 15);
      }
      o = 0;
      while (o < 81)
      {
        tnum = enemy_on_tile(o);
        if (tnum != -1)
        {
          draw_sample_enemy(enemy[tnum].type, tile[o].x, tile[o].y);
        }
        o++;
      }
      break;
  }
  doparticles();
}





void logic()
{
  clear(back_buffer);
  if (initialized == 0)
  {
    levelinit();
    draw_all();                                               //this just makes it fade
    blit(back_buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);//in to the main game and
    fade_in((RGB *)palette[0].dat, 5);                        //not the menu screen
    counter = 0;
    initialized = 1;
  }
  else
  {
    if (died == 0)
    {
      if ((key[KEY_ESC]) && (keydelay <= 0))  //as explained in query_exit(), keydelay
      {                                       //is a global variable used for keeping
        query_exit();                         //the game from reading the escape key
      }                                       //right after the player returns to the
      else if (key[KEY_ESC])                  //game from query_exit()
      {
        keydelay--;
      }
      if (player[0].trapped == 0)
      {
        check_keys();  //if you've been killed or trapped you can't move or cast spells
      }
    }
    do_ai();
    draw_all();
    if (check_for_loss() != 1)
    {
      check_for_win();
    }
    gametime++;  //gametime is the counter for the time bonus, not for the fps
  }
}





void levelinit()
{
  int w;
  if (level < 101)  //if not a warp level
  {
    switch(level)
    {
      case 1:
        create_enemy(1, 16);
        break;
      case 2:
        create_enemy(1, 16);
        create_enemy(1, 11);
        create_enemy(1, 22);
        break;
      case 3:
        create_enemy(1, 16);
        create_enemy(1, 29);
        create_enemy(1, 7);
        create_enemy(1, 53);
        create_enemy(1, 5);
        break;
      case 4:
        create_enemy(2, 16);
        break;
      case 5:
        create_enemy(2, 16);
        create_enemy(1, 29);
        create_enemy(1, 7);
        break;
      case 6:
        create_enemy(2, 16);
        create_enemy(2, 29);
        create_enemy(2, 7);
        break;
      case 7:
        create_enemy(2, 16);
        create_enemy(2, 29);
        create_enemy(2, 7);
        create_enemy(2, 23);
        create_enemy(1, 53);
        create_enemy(1, 5);
        break;
      case 8:
        create_enemy(4, 17);
        break;
      case 9:
        create_enemy(1, 30);
        create_enemy(1, 17);
        create_enemy(4, 31);
        break;
      case 10:
        create_enemy(4, 66);
        create_enemy(2, 37);
        create_enemy(2, 22);
        create_enemy(2, 11);
        create_enemy(2, 4);
        create_enemy(1, 9);
        break;
      case 11:
        create_enemy(4, 2);
        create_enemy(4, 7);
        create_enemy(1, 16);
        create_enemy(4, 29);
        create_enemy(4, 45);
        break;
      case 12:
        create_enemy(5, 16);
        break;
      case 13:
        create_enemy(5, 16);
        create_enemy(1, 21);
        create_enemy(1, 3);
        break;
      case 14:
        create_enemy(5, 0);
        create_enemy(5, 36);
        break;
      case 15:
        create_enemy(5, 16);
        create_enemy(4, 29);
        create_enemy(4, 7);
        break;
      case 16:
        create_enemy(5, 0);
        create_enemy(5, 23);
        create_enemy(5, 36);
        create_enemy(2, 22);
        create_enemy(2, 11);
        break;
      case 17:
        create_enemy(2, 36);
        create_enemy(2, 28);
        create_enemy(2, 21);
        create_enemy(2, 15);
        create_enemy(2, 10);
        create_enemy(2, 6);
        create_enemy(2, 3);
        create_enemy(2, 1);
        create_enemy(2, 0);
        break;
      case 18:
        create_enemy(5, 37);
        create_enemy(5, 4);
        create_enemy(1, 15);
        create_enemy(1, 6);
        create_enemy(4, 16);
        create_enemy(2, 11);
        create_enemy(2, 22);
        break;
      case 19:
        create_enemy(6, 16);
        break;
      case 20:
        create_enemy(6, 11);
        create_enemy(6, 22);
        break;
      case 21:
        create_enemy(6, 16);
        create_enemy(1, 4);
        create_enemy(1, 37);
        create_enemy(2, 30);
        create_enemy(2, 23);
        create_enemy(2, 17);
        create_enemy(1, 31);
        break;
      case 22:
        create_enemy(6, 60);
        create_enemy(6, 9);
        create_enemy(4, 22);
        create_enemy(4, 11);
        create_enemy(5, 21);
        create_enemy(5, 3);
        break;
      default:
        w = check_highscores();
        if (w == -1)
        {
          gamestatus = 0;
        }
        else if (w > -1)
        {
          gamestatus = 5;
        }
        break;
    }
  }
  else  //if it is a warp level
  {
    create_enemy(1, 67);
    create_enemy(1, 26);
    create_enemy(1, 16);
    create_enemy(2, 47);
    create_enemy(2, 18);
    drop_item(77, 0);
    drop_item(74, 1);
    drop_item(70, 4);
    drop_item(65, 2);
    drop_item(59, 3);
  }
}





void init_particle(int type, int x, int y, int color)
{
  init_particle(type, x, y, color, 0, 0);
} //In particles 1-6, the xv and yv are determined in the init, but
  //for particle 7, the xv and yv are given by an outside function.




void init_particle(int type, int x, int y, int color, int xv, int yv)
{
  int num = find_free();
  int temp;
  if (num != -1)
  {
    particle[num].exist = 1;
    particle[num].x = x * 10;  //these are multiplied by ten and later
    particle[num].y = y * 10;  //divided by ten to give the illusion of floats
    particle[num].type = type;
    switch (type)
    {
      case 1:
        temp = random()%2;
        if (temp == 0)
        {
          temp = -1;
        }
        particle[num].life = temp;  //"life" not used as life
        particle[num].xv = 2 * 10;
        particle[num].yv = ((random()%4) + 2) * 10;
        particle[num].color = (random()%10) + 21;  //different shades of grey
        break;
      case 2:
        particle[num].life = 10;
        particle[num].xv = (random()%15);
        temp = random()%2;
        if (temp == 0)
        {
          temp = -1;
        }
        particle[num].xv = particle[num].xv * temp;
        particle[num].yv = (random()%15);
        temp = random()%2;
        if (temp == 0)
        {
          temp = -1;
        }
        particle[num].yv = particle[num].yv * temp;
        particle[num].color = color;
        break;
      case 3:
        temp = random()%2;
        if (temp == 0)
        {
          temp = -1;
        }
        particle[num].life = 20;
        particle[num].temp = temp;
        particle[num].xv = 2 * 10;
        particle[num].yv = ((random()%4) + 2) * 10;
        particle[num].color = (random()%10) + 21;
        break;
      case 4:
        particle[num].life = 15;
        particle[num].xv = 0;
        particle[num].yv = 0;
        particle[num].color = color;
        break;
      case 5:
        particle[num].life = 20;
        temp = random()%2;
        if (temp == 0)
        {
          temp = -1;
        }
        particle[num].x += (temp * random()%200);
        particle[num].xv = temp * -5;
        temp = random()%2;
        if (temp == 0)
        {
          temp = -1;
        }
        particle[num].y += (temp * random()%200);
        particle[num].yv = temp * -5;
        particle[num].color = color;
        break;
      case 6:
        particle[num].life = 4 + random()%4;
        particle[num].xv = 0;
        particle[num].yv = 0;
        particle[num].color = color;
        break;
      case 7:
        particle[num].life = 8;
        particle[num].xv = xv * 10; //more multiplication by 10 stuff
        particle[num].yv = yv * 10;
        particle[num].color = color;
        break;
      case 8:
        particle[num].life = 8;
        particle[num].temp = 8;
        particle[num].xv = 0;
        particle[num].yv = 0;
        particle[num].color = 40;
        break;
      case 9:
        particle[num].life = 25;
        particle[num].temp = random()%360;
        particle[num].xv = 0;
        particle[num].yv = -35 + random()%20;  //would be -3.5 + random()%2, but the
        break;                                 //hackish x10 float-simulator rears
      case 10:                                 //its ugly head once more
        particle[num].life = 20;
        particle[num].temp = 12;
        particle[num].xv = 0;
        particle[num].yv = 0;
        particle[num].color = color;
        break;
    }
    if (num > final_particle)
    {
      final_particle = num + 1;
    }
  }
}





void doparticles()
{
  int a = 0;
  int x = 0;
  int y = 0;
  float temp;
  while (a < final_particle)//PARTICLENUM)
  {
    if (particle[a].exist == 1)
    {
      switch (particle[a].type)
      {
        case 1:
          particle[a].xv += particle[a].life * 10;
          if (particle[a].xv > 3 * 10)  //More of that * 10 stuff...
          {
            particle[a].life = -1;
          }
          else if (particle[a].xv < -3 * 10)  //Still more of the same
          {
            particle[a].life = 1;
          }
          particle[a].x += particle[a].xv;
          particle[a].y -= particle[a].yv;
          if (particle[a].y < 0)
          {
            particle[a].exist = 0;
          }
          x = particle[a].x / 10;  //Now it's divided, which lets me
          y = particle[a].y / 10;  //avoid floats that never seem to work
          putpixel(back_buffer, x, y, particle[a].color);
          putpixel(back_buffer, x + 1, y + 1, particle[a].color);
          putpixel(back_buffer, x, y, particle[a].color);
          putpixel(back_buffer, x + 1, y + 1, particle[a].color);
          break;
        case 2:
          particle[a].x += particle[a].xv;
          particle[a].y -= (particle[a].yv);
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          x = particle[a].x / 10;  //Same division by ten
          y = particle[a].y / 10;  //stuff over here
          putpixel(back_buffer, x, y, particle[a].color);
          break;
        case 3:
          particle[a].xv += particle[a].temp * 10;
          if (particle[a].xv > 3 * 10)
          {
            particle[a].temp = -1;
          }
          else if (particle[a].xv < -3 * 10)
          {
            particle[a].temp = 1;
          }
          particle[a].x += particle[a].xv;
          particle[a].y -= particle[a].yv;
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          x = particle[a].x / 10;
          y = particle[a].y / 10;
          putpixel(back_buffer, x, y, particle[a].color);
          putpixel(back_buffer, x + 1, y + 1, particle[a].color);
          putpixel(back_buffer, x, y, particle[a].color);
          putpixel(back_buffer, x + 1, y + 1, particle[a].color);
          break;
        case 4:
          int u;
          int j;
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          j = 0;
          while (j < 20)
          {
            u = find_free();
            x = particle[a].x / 10;
            y = particle[a].y / 10;
            init_particle(2, x, y, particle[a].color);
            particle[u].life = 5;  //a hack to reduce the life of the new particle
            j++;
          }
          break;
        case 5:
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          particle[a].x += particle[a].xv;
          particle[a].y += particle[a].yv;
          x = particle[a].x / 10;
          y = particle[a].y / 10;
          putpixel(back_buffer, x, y, particle[a].color);
          break;
        case 6:
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          x = particle[a].x / 10;
          y = particle[a].y / 10;
          putpixel(back_buffer, x, y, particle[a].color);
          break;
        case 7:
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          x = (particle[a].x += particle[a].xv) / 10;  //more division by 10 stuff
          y = (particle[a].y += particle[a].yv) / 10;
          putpixel(back_buffer, x, y, particle[a].color);
          break;
        case 8:
          if (particle[a].temp - particle[a].life > 3)
          {
            particle[a].color = 43;
            if (particle[a].temp - particle[a].life > 5)
            {
              particle[a].color = 44;
            }
          }
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          x = particle[a].x / 10;
          y = particle[a].y / 10;
          putpixel(back_buffer, x, y, particle[a].color);
          break;
        case 9:
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          particle[a].x += particle[a].xv;
          particle[a].y += particle[a].yv;
          x = particle[a].x / 10;
          y = particle[a].y / 10;
          particle[a].temp += 12;
          if (particle[a].temp >= 360)
          {
            particle[a].temp -= 360;
          }
          rotate_sprite(back_buffer, (BITMAP *)misc[1].dat, x, y, itofix(particle[a].temp));
          break;
        case 10:
          particle[a].life--;
          if (particle[a].life <= 0)
          {
            particle[a].exist = 0;
          }
          particle[a].temp += 12;
          if (particle[a].temp < 240)
          {
            ellipsefill(back_buffer, 320, 240, particle[a].temp * 2, particle[a].temp, particle[a].color);
          }
          else
          {
            clear_to_color(back_buffer, 15);
          }
          break;
      }
      if (particle[a].exist == 0)
      {
        if (a == final_particle - 1) //if a was the highest used particle
        {
          final_particle--;  //take the particles looped through down a notch
          while (particle[final_particle - 1].exist == 0)
          {                    //keeps reducing the particles looped through
            final_particle--;  //until it hits an on-screen particle
          }
        }
      }
    }
    a++;
  }
}





int find_free()
{
  int search = 0;
  int found = 0;
  while ((search < PARTICLENUM) && (found == 0))
  {
    if (particle[search].exist == 0)
    {
      found = 1;
    }
    else
    {
      search++;
    }
  }
  if (search < PARTICLENUM)
  {
    return search;
  }
  else
  {
    return -1;
  }
}





void show_highscores()
{
  int q = 1;
  int i;
  char name[15];
  int points;
  clear(back_buffer);
  textout_centre(back_buffer, (FONT *)fonts[0].dat, "Top Ten Mages", 320, 40, 15);
  i = 0;
  while (i < 10)
  {
    textout(back_buffer, font, names[i].c_str(), 190, 100 + (i * 25), 15);
    sprintf(msg, "%-1d", oldscores[i]);
    textout_right(back_buffer, font, msg, 450, 100 + (i * 25), 15);
    i++;
  }
  textout_centre(back_buffer, font, "Press R and S together to reset the table", 320, 450, 2);
  textout_centre(back_buffer, font, "Press ESC to return to the main menu", 320, 460, 2);
  if ((key[KEY_ESC]) && (keydelay <= 0))
  {
    gamestatus = 0;
  }
  else if (key[KEY_ESC])
  {
    keydelay--;  //check the description of keydelay in the check_keys()
  }
  if ((key[KEY_R]) && (key[KEY_S]))
  {
    reset_table();
  }
}





void reset_table()
{
  int i;
  i = 0;
  while (i < 10)
  {
    names[i] = "Kent";
    oldscores[i] = 10000 - (1000 * i);
    i++;
  }
}





int check_highscores()
{
  int i;
  int done = 0;
  i = 0;
  while (i < 10)
  {
    if (score >= oldscores[i])
    {
      done = 1;
      return i;
    }
    i++;
  }
  if (done == 0)
  {
    return -1;
  }
}





void post_score(int loc, string entry)
{
  int i;
  i = 9;
  while (i > loc)
  {
    names[i] = names[i - 1];
    oldscores[i] = oldscores[i - 1];
    i--;
  }
  names[loc] = entry;
  oldscores[loc] = score;
  keydelay = 5;  //check the description of keydelay in the check_keys()
}





void enter_highscore()
{
  clear(back_buffer);
  clear_keybuf();
  int ok = 0;
  int q;
  char name[10];
  q = check_highscores();
  textout_centre(back_buffer, (FONT *)fonts[0].dat, "New High Score!", 320, 10, 15);
  textout_centre(back_buffer, font, "Enter Your name:", 320, 80, 15);
  blit(back_buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
//-------------------------
  int x = 250;  //This code was blatantly ripped from a post on the Allegro.cc forums.
  int y = 100;  //It was posted by Chris Barry (23yrold3yrold), and then tweaked for
  int w = 140;  //my use (by me, of course).
  int h = 15;
  int leng = 10;
  int i = 0;
  int k;
  BITMAP *bmp;
  bmp = create_bitmap(w, h);
  clear(bmp);
  rect(bmp, 0, 0, w - 1, h - 1, 2);
  name[0] = '\0';
  rectfill(bmp, 1, 1, w - 2, h - 2, makecol(50, 50, 50));
  textprintf(bmp, font, 2, 2, makecol(0, 200, 0), "%s", name);
  draw_sprite(screen, bmp, x, y);
  clear_keybuf();
  while (ok == 0)
  {
    k = readkey();
    if (((k >> 8) == KEY_ENTER))
    {
      if (i > 0)
      {
        ok = 1;
      }
    }
    else if (((k >> 8) == KEY_BACKSPACE) && i > 0)
    {
      i--;
      name[i] = '\0';
    }
    else if (((k >> 8) == KEY_BACKSPACE) && i <= 0)
    {
      //do nothing
    }    
    else if (((k >> 8) == KEY_ESC))
    {
      if (i > 0)
      {
        ok = 1;
      }
    }
    else if (((k >> 8) == KEY_TAB) || ((k >> 8) == KEY_SPACE))
    {
      //do nothing
    }
    else
    {
      if (i < leng)
      {
        name[i] = k & 0xFF;
        name[i + 1] = '\0';
        i++;
      }
    }
    rectfill(bmp, 1, 1, w - 2, h - 2, makecol(50, 50, 50));
    textprintf(bmp, font, 2, 2, makecol(0, 200, 0), "%s", name);
    draw_sprite(screen, bmp, x, y);
  }
//-------------------
  textout_centre(screen, font, "Adding Score...", 320, 460, 15);
  post_score(q, name);
  gamestatus = 4;
  counter = 0;
}





void draw_board()
{
  static int firsttime = 1;
  int pos = 1;
  int count = 1;
  int layer = 1;
  int i = 0;
  int layerc = 1;
  int temp;
  int max;
  int ptrapped;
  while (i < 81)
  {
    if ((tile[i].status == -1) || (tile[i].status == -2))  //if it's either normal
    {                                                      //or an untriggered trap
      draw_sprite(back_buffer, (BITMAP *)tiles[0].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
    }
    else  //if there is an active spell on it
    {
      switch(tile[i].status)
      {
        case 5:  //jab spell
          draw_sprite(back_buffer, (BITMAP *)jab[tile[i].frame / 5].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          tile[i].frame++;
          if (tile[i].frame > 14)
          {
            tile[i].frame = 0;
            tile[i].status = -1;
          }
          break;
        case 9:  //spike trap
          draw_sprite(back_buffer, (BITMAP *)spike[tile[i].frame / 5].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          tile[i].frame++;
          if (tile[i].frame > 19)
          {
            tile[i].frame = 0;
            tile[i].status = -1;
          }
          break;
        case 10:  //needle spell
          draw_sprite(back_buffer, (BITMAP *)needle[tile[i].frame / 5].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          tile[i].frame++;
          if (tile[i].frame > 14)
          {
            tile[i].frame = 0;
            tile[i].status = -1;
          }
          break;
        case 12:  //lightning rod trap
        case 25:  //plasmic trap
          draw_sprite(back_buffer, (BITMAP *)tiles[0].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          temp = find_free_graphic();  // /\ This trap doesn't replace the tile with
          if (temp != -1)              // || its own sprite, so it draws a normal tile
          {                            //    here. Otherwise the tile would be undrawn
            graphic[temp].exist = 1;
            graphic[temp].max = 1;
            graphic[temp].life = 20;
            graphic[temp].x = tile[i].x + 19;
            graphic[temp].y = tile[i].y - 75;
            graphic[temp].type = 11;
          }
          temp = 0;
          while (temp < 30)
          {
            init_particle(2, tile[i].x + 25, tile[i].y + 19, 15);
            init_particle(2, tile[i].x + 25, tile[i].y + 19, 53);  //light blue
            temp++;
          }
          tile[i].status = -1;
          break;
        case 14:  //firepit trap
          draw_sprite(back_buffer, (BITMAP *)firepit[tile[i].frame / 5].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          tile[i].frame++;
          if (tile[i].frame > 19)
          {
            tile[i].frame = 0;
            tile[i].status = -1;
          }
          break;
        case 31:  //fireshock trap
          temp = find_free_graphic();  //it basically combines Lightning-Rod
          if (temp != -1)              //and Firepit's effects
          {
            graphic[temp].exist = 1;
            graphic[temp].max = 1;
            graphic[temp].life = 20;
            graphic[temp].x = tile[i].x + 19;
            graphic[temp].y = tile[i].y - 75;
            graphic[temp].type = 11;
          }
          temp = 0;
          while (temp < 15)
          {
            init_particle(2, tile[i].x + 25, tile[i].y + 19, 15);
            init_particle(2, tile[i].x + 25, tile[i].y + 19, 53);  //light blue
            temp++;
          }
          draw_sprite(back_buffer, (BITMAP *)firepit[tile[i].frame / 5].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          tile[i].frame++;
          if (tile[i].frame > 19)
          {
            tile[i].frame = 0;
            tile[i].status = -1;
          }
          break;
        case 18:  //ice trap
        case 22:  //arctic trap
        case 23:  //subzero trap
        case 33:  //encase trap
        case 36:  //amber trap
          if (tile[i].status == 18)
          {
            max = 1500;
          }
          else if (tile[i].status == 22)
          {
            max = 2000;
          }
          else if (tile[i].status == 23)
          {
            max = 2500;
          }
          else if (tile[i].status == 33)
          {
            max = 1800;
          }
          else if (tile[i].status == 36)
          {
            max = 3000;
          }
          temp = enemy_on_tile(i);
          if (temp == -1)  //if the trap killed the enemy or it wasn't an enemy that
          {                //was trapped
            temp = player_on_tile(i);
            if (temp == -1)  //if there wasn't a player or ally there either
            {
              tile[i].frame = max + 1;  //destroy the trap
            }
            else
            {
              ptrapped = 1;
              player[temp].trapped = 1;
            }
          }
          else
          {
            enemy[temp].trapped = 1;
            ptrapped = 0;
          }
          draw_sprite(back_buffer, (BITMAP *)tiles[0].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          tile[i].frame++;
          if (tile[i].frame > max)  //the trap faded
          {
            tile[i].frame = 0;
            tile[i].status = -1;
            if (ptrapped == 0)
            {
              enemy[temp].trapped = 0;
            }
            else
            {
              player[temp].trapped = 0;
            }
          }
          break;
        case 32:  //earthquake spell
          draw_sprite(back_buffer, (BITMAP *)quake[tile[i].frame / 5].dat, 270 - (23 * count) + (46 * pos), 20 + (layer * 23));
          tile[i].frame++;
          if (tile[i].frame > 34)
          {
            tile[i].frame = 0;
            tile[i].status = -1;
          }
          break;
        case 34:  //world-tremor spell
          tile[i].frame++;
          if (tile[i].frame%2 == 0)  //this shakes the tile back and forth
          {
            temp = 1;
          }
          else
          {
            temp = -1;
          }
          draw_sprite(back_buffer, (BITMAP *)tiles[0].dat, 270 - (23 * count) + (46 * pos) + temp, 20 + (layer * 23));
          if (tile[i].frame > 35)
          {
            tile[i].frame = 0;
            tile[i].status = -1;
          }
          break;
      }
    }
    if (firsttime == 1)
    {
      tile[i].x = 270 - (23 * count) + (46 * pos);
      tile[i].y = 20 + (layer * 23);
      if (i == 80)
      {
        firsttime = 0;
      }
    }
    pos++;
    if (pos > count)
    {
      pos -= (count * 1);
      count += layerc;
      layer++;
      if (count > 8)
      {
        layerc = -1;
      }
    }
    i++;
  }
}





void draw_trap(mover current)
{
  switch(tile[current.location].status)
  {
    case 18:  //ice
    case 22:  //arctic
    case 23:  //subzero
      draw_trans_sprite(back_buffer, (BITMAP *)ice[0].dat, tile[current.location].x, tile[current.location].y - 10);
      break;
    case 33:  //encase
      draw_sprite(back_buffer, (BITMAP *)encase[0].dat, tile[current.location].x + 10, tile[current.location].y - 5);
      break;
    case 36:  //amber
      draw_trans_sprite(back_buffer, (BITMAP *)amber[0].dat, tile[current.location].x + 10, tile[current.location].y - 5);
      break;
  }
}





void draw_shield(int shieldtype, int x, int y)
{
  switch(shieldtype)
  {
    case 1:
      draw_sprite(back_buffer, (BITMAP *)shield[0].dat, x + 8, y - 4);
      init_particle(2, x + 25, y + 18, 43);
      break;
    case 2:
      draw_sprite(back_buffer, (BITMAP *)barrier[0].dat, x + 8, y - 4);
      init_particle(2, x + 25, y + 18, 47);
      break;
    case 3:
      draw_sprite(back_buffer, (BITMAP *)power[0].dat, x + 8, y - 4);
      init_particle(2, x + 25, y + 18, 15);
      break;
  }
}





void draw_all()
{
  draw_board();
  draw_stats();
  int a = 0;
  int e;
  while (a < 81)
  {
    if (tile[a].object != -1)
    {
      draw_object(a, tile[a].object);
    }
    if ((player[0].oldlocation == a) && (player[0].exist == 1))
    {
      player[0] = calc_mover(player[0]);
      if (tile[a].object != -1)
      {
        grab_item(tile[a].object);
        tile[a].object = -1;
      }
    }
    else
    {
      e = 0;
      while (e < 15)
      {
        if ((enemy[e].exist == 1) && (enemy[e].oldlocation == a))
        {
          enemy[e] = calc_mover(enemy[e]);
          if (tile[a].object != -1)
          {
            tile[a].object = -1;  //if an enemy lands on a tile with
          }                       //an item, then the item disappears
        }
        e++;
      }
    }
    a++;
  }
  draw_spells();
  doparticles();
}





void draw_stats()
{
  int x = 0;
  int y = 0;
  int z = 0;
  while (z < player[0].health)
  {
    rectfill(back_buffer, 500 + (x * 10), 350 + (y * 10), 505 + (x * 10), 355 + (y * 10), 4);
    x++;
    if (x > 9)
    {
      x = 0;
      y++;
    }
    z++;
  }
  sprintf(msg, "%-1d", score);
  textout_right(back_buffer, (FONT *)fonts[2].dat, msg, 200, 20, 11);
  sprintf(msg, "%-1d", level);
  textout_right(back_buffer, (FONT *)fonts[2].dat, msg, 200, 40, 10);
}





void draw_object(int tilenum, int type)
{
  draw_sprite(back_buffer, (BITMAP *)orbs[type].dat, tile[tilenum].x + 17, tile[tilenum].y + 18);
}  //this used to be longer and actually be worth having a function for it





mover calc_mover(mover current)
{
  int offset_x = 0;
  int offset_y = 0;
  int t;
  int mcount = 0;  //only useful in the one frame after a character stops moving
  if (current.moving != -1)
  {
    current.movecount++;
    mcount = current.movecount;  //normally mcount does nothing, but it has a use later
    switch (current.moving)
    {
      case 1:
        if (current.attacked == 1)
        {
          offset_x = 22 - current.movecount;
          offset_y = -1 * (22 - current.movecount);
        }
        else
        {
          offset_x = -1 * current.movecount;
          offset_y = current.movecount;
          if (current.movecount > 11)
          {
            current.location = tile[current.oldlocation].up;
          }
        }
        break;
      case 2:
        if (current.attacked == 1)
        {
          offset_x = -1 * (22 - current.movecount);
          offset_y = 22 - current.movecount;
        }
        else
        {
          offset_x = current.movecount;
          offset_y = -1 * current.movecount;
          if (current.movecount > 11)
          {
            current.location = tile[current.oldlocation].down;
          }
        }
        break;
      case 3:
        if (current.attacked == 1)
        {
          offset_x = 22 - current.movecount;
          offset_y = 22 - current.movecount;
        }
        else
        {
          offset_x = -1 * current.movecount;
          offset_y = -1 * current.movecount;
          if (current.movecount > 11)
          {
            current.location = tile[current.oldlocation].left;
          }
        }
        break;
      case 4:
        if (current.attacked == 1)
        {
          offset_x = -1 * (22 - current.movecount);
          offset_y = -1 * (22 - current.movecount);
        }
        else
        {
          offset_x = current.movecount;
          offset_y = current.movecount;
          if (current.movecount > 11)
          {
            current.location = tile[current.oldlocation].right;
          }
        }
        break;
    }
    if (current.type != 0)  //the current mover is not the player
    {
      if (current.location == player[0].location)  //the enemy is on the player's tile
      {
        if ((player[0].shielded != 1) && (player[0].shielded != 3))
        {  //see shielded definitions in the mover struct for explanation
          player[0].health -= current.power;
        }
        current.attacked = 1;
        current.moving = invert_move(current.moving);
        current.location = current.oldlocation;
        t = 0;
        while (t < 50)
        {
          init_particle(2, tile[player[0].location].x + 25, tile[player[0].location].y + 15, 12);
          t++;
        }
      }
    }
    if (current.movecount > 22)
    {
      mcount = current.movecount;  //if the character wants to move to another
      current.movecount = 0;       //tile, mcount will keep the animation fluid
      current.moving = -1;
      current.oldlocation = current.location;
      current.attacked = 0;
      offset_x = 0;
      offset_y = 0;
      if (tile[current.location].status == -2)  //if the mover just walked on a trap
      {
        int l = current.location;
        tile[l].status = tile[l].spell;  //activates trap
        current = deal_damage(current, archive[tile[l].spell - 1].power, archive[tile[l].spell - 1].element);
        current.trapped = 1;
      }
    }
  }
  if ((current.type != 0) || (died == 0))
  {  //doesn't draw the player if he's dead
    draw_mover(current.oldlocation, current.type, offset_x, offset_y, current.direction, mcount);
  }
  if (current.shielded != 0)
  {
    draw_shield(current.shielded, tile[current.oldlocation].x + offset_x, tile[current.oldlocation].y - offset_y);
    current.shieldcount--;
    if (current.shieldcount <= 0)
    {
      current.shielded = 0;
    }
  }
  if (current.trapped == 1)
  {
    draw_trap(current);
  }
  return current;
}





void draw_mover(int tilenum, int type, int offset_x, int offset_y, int direction, int anim)
{
  int xthing = tile[tilenum].x + offset_x;
  int ything = tile[tilenum].y - offset_y;
  int walking;
  if (anim == 0)
  {
    walking = 0;
  }
  else
  {
    walking = 1;
  }
  switch(type)
  {
    case 0:
      while ((anim / 4) > 5)  //the "/4" slows down the animation, this character
      {                       //has 6 frames of motion
        anim -= 24;  //would be 6 except for the "/4"
      }
      anim = anim / 4;  //more "/4" stuff
      switch(direction)
      {
        case 1:
          if (walking == 0)  //if the player isn't moving
          {
            draw_sprite(back_buffer, (BITMAP *)mage[0].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)magewalk1[anim].dat, xthing, ything - 10);
          }
          break;
        case 2:
          if (walking == 0)
          {
            draw_sprite(back_buffer, (BITMAP *)magewalk2[1].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)magewalk2[anim].dat, xthing, ything - 10);
          }
          break;
        case 3:
          if (walking == 0)
          {
            draw_sprite(back_buffer, (BITMAP *)magewalk3[1].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)magewalk3[anim].dat, xthing, ything - 10);
          }
          break;
        case 4:
          if (walking == 0)
          {
            draw_sprite(back_buffer, (BITMAP *)mage[1].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)magewalk4[anim].dat, xthing, ything - 10);
          }
          break;
      }
      break;
    case 1:
      draw_sprite(back_buffer, (BITMAP *)slime[0].dat, xthing, ything);
      break;
    case 2:
      while ((anim / 4) > 5)  //the "/4" slows down the animation, this character
      {                       //has 6 frames of motion
        anim -= 24;  //would be 6 except for the "/4"
      }
      anim = anim / 4;  //more "/4" stuff
      switch(direction)
      {
        case 1:
          if (walking == 0)
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk1[2].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk1[anim].dat, xthing, ything - 10);
          }
          break;
        case 2:
          if (walking == 0)
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk2[2].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk2[anim].dat, xthing, ything - 10);
          }
          break;
        case 3:
          if (walking == 0)
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk3[4].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk3[anim].dat, xthing, ything - 10);
          }
          break;
        case 4:
          if (walking == 0)
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk4[2].dat, xthing, ything - 10);
          }
          else
          {
            draw_sprite(back_buffer, (BITMAP *)gobwalk4[anim].dat, xthing, ything - 10);
          }
          break;
      }
      break;
    case 3:
      if (tilenum != -2)  //mole isn't hidden
      {
        if (direction < 27)  //direction is used as temp for the mole
        {
          draw_sprite(back_buffer, (BITMAP *)molepop[direction / 3].dat, xthing + 7, ything + 2);
        }
        else if (direction > 123)  //123 is 150 - 28
        {
          draw_sprite(back_buffer, (BITMAP *)molepop[(150 - direction) / 3].dat, xthing + 7, ything + 2);
        }
        else
        {
          draw_sprite(back_buffer, (BITMAP *)mole[0].dat, xthing + 7, ything + 2);
        }
      }
      break;
    case 4:
      if ((anim / 2) > 5)  //the "/4" slows down the animation
      {
        anim -= 12;  //would be 6 except for the "/4"
        anim = 12 - anim;
      }
      anim /= 2;  //more "/4" stuff
      if (anim >= 5)  //it shouldn't exceed 5, but if it does the '>'
      {               //will keep the game from crashing
        draw_sprite(back_buffer, (BITMAP *)clay[0].dat, xthing + 8, ything + 17);
      }
      else
      {
        switch(direction)
        {
          case 1:
            draw_sprite_h_flip(back_buffer, (BITMAP *)clayur[anim].dat, xthing, ything - 10);
            break;
          case 2:
            draw_sprite_h_flip(back_buffer, (BITMAP *)claydl[anim].dat, xthing, ything - 10);
            break;
          case 3:
            draw_sprite(back_buffer, (BITMAP *)claydl[anim].dat, xthing, ything - 10);
            break;
          case 4:
            draw_sprite(back_buffer, (BITMAP *)clayur[anim].dat, xthing, ything - 10);
            break;
        }
      }
      break;
    case 5:
      switch(direction)
      {
        case 1:
          draw_trans_sprite(back_buffer, (BITMAP *)phantom[0].dat, xthing + 20, ything + 8);
          break;
        case 2:
          draw_trans_sprite(back_buffer, (BITMAP *)phantom[2].dat, xthing + 20, ything + 8);
          break;
        case 3:
          draw_trans_sprite(back_buffer, (BITMAP *)phantom[3].dat, xthing + 20, ything + 8);
          break;
        case 4:
          draw_trans_sprite(back_buffer, (BITMAP *)phantom[1].dat, xthing + 20, ything + 8);
          break;
      }
      break;
    case 6:
      while ((anim / 4) > 3)  //the "/4" slows down the animation, this character
      {                       //has 4 frames of motion
        anim -= 16;  //would be 4 except for the "/4"
      }
      anim = anim / 4;  //more "/4" stuff
      switch(direction)
      {
        case 1:
          draw_sprite(back_buffer, (BITMAP *)imp1[anim].dat, xthing + 3, ything - 3);
          break;
        case 2:
          draw_sprite(back_buffer, (BITMAP *)imp2[anim].dat, xthing + 3, ything - 3);
          break;
        case 3:
          draw_sprite(back_buffer, (BITMAP *)imp3[anim].dat, xthing + 3, ything - 3);
          break;
        case 4:
          draw_sprite(back_buffer, (BITMAP *)imp4[anim].dat, xthing + 3, ything - 3);
          break;
      }
      break;
  }
}





void check_keys()
{
  if (player[0].movecount == 0)
  {
    if (key[KEY_UP])
    {
      player[0].direction = 1;
      if (tile[player[0].location].up != -1)
      {
        player[0].moving = check_move(player[0].oldlocation, 1);
      }
      player[0].oldlocation = player[0].location;
    }
    else if (key[KEY_DOWN])
    {
      player[0].direction = 2;
      if (tile[player[0].location].down != -1)
      {
        player[0].moving = check_move(player[0].oldlocation, 2);
      }
      player[0].oldlocation = player[0].location;
    }
    else if (key[KEY_LEFT])
    {
      player[0].direction = 3;
      if (tile[player[0].location].left != -1)
      {
        player[0].moving = check_move(player[0].oldlocation, 3);
      }
      player[0].oldlocation = player[0].location;
    }
    else if (key[KEY_RIGHT])
    {
      player[0].direction = 4;
      if (tile[player[0].location].right != -1)
      {
        player[0].moving = check_move(player[0].oldlocation, 4);
      }
      player[0].oldlocation = player[0].location;
    }
    else if (key[KEY_1])
    {
      if ((spell[0].active == 1) && (spell[0].percent == 100))
      {
        player[0] = cast_spell(player[0], 0);
      }
    }
    else if (key[KEY_2])
    {
      if ((spell[1].active == 1) && (spell[1].percent == 100))
      {
        player[0] = cast_spell(player[0], 1);
      }
    }
    else if (key[KEY_3])
    {
      if ((spell[2].active == 1) && (spell[2].percent == 100))
      {
        player[0] = cast_spell(player[0], 2);
      }
    }
    else if (key[KEY_4])
    {
      if ((spell[3].active == 1) && (spell[3].percent == 100))
      {
        player[0] = cast_spell(player[0], 3);
      }
    }
    else if (key[KEY_5])
    {
      if ((spell[4].active == 1) && (spell[4].percent == 100))
      {
        player[0] = cast_spell(player[0], 4);
      }
    }
  }
}





void do_ai()
{
  int i = 0;
  while (i < 15)
  {
    if ((enemy[i].exist == 1) && (enemy[i].trapped == 0))
    {
      enemy[i] = calc_mover_ai(enemy[i]);
    }
    i++;
  }
}





mover calc_mover_ai(mover current)
{
  int i;
  int ok;
  switch (current.type)
  {
    case 1:
    case 2:
    case 4:
    case 5:
    case 6:  //same actions for slime, goblin, clayman, phantom, and imp
      if ((current.temp > ((current.speed / 2) + ((random()%current.speed) * 2))) && (current.moving == -1))
      {
        current.temp = 0;
        current.moving = -1;
        while (current.moving == -1)
        {
          current.moving = target_player(current);
          current.moving = check_move(current.oldlocation, current.moving);
        }
        current.direction = current.moving;
      }
      current.temp++;
      break;
    case 3:  //mole
      if (current.location != -2)  //if the mole is not off the board (in hiding)
      {
        current.temp++;
        if (current.temp == 50)  //a little after the mole has fully popped up
        {
          current.direction = 2;  //hack to make cast_spell work
          current = cast_enemy_spell(current, 32);  //cast earthquake
          current.direction = 28;
        }
        if (current.temp == 150)  //after 3 seconds of being up
        {
          current.location = -2;  //go off the board
          current.temp = 0;
        }
      }
      else
      {
        current.temp++;
        if (current.temp >= 150 + rand()%200)  //after 3 or so seconds of being hidden
        {
          ok = 0;
          while (ok == 0)
          {
            i = random()%81;
            if (player_on_tile(i) + enemy_on_tile(i) == -2)  //both functions return -1
            {
              ok = 1;
            }
          }
          current.location = i;
          current.temp = 0;
        }
      }
      current.direction = current.temp;  //just a hack to make it work with draw_mover
      current.oldlocation = current.location;
      break;
  }
  return current;
}





int target_player(mover attacker)
{
  int x;
  int y;
  int randmove = random()%4;
  if (randmove != 1)
  {
    if (tile[player[0].location].x > tile[attacker.location].x)
    {
      x = 1;
    }
    else
    {
      x = -1;
    }
    if (tile[player[0].location].y < tile[attacker.location].y)
    {
      y = 1;
    }
    else
    {
      y = -1;
    }
    if ((x == -1) && (y == 1))
    {
      return 1;
    }
    else if ((x == 1) && (y == -1))
    {
      return 2;
    }
    else if ((x == -1) && (y == -1))
    {
      return 3;
    }
    else
    {
      return 4;
    }
  }
  else
  {
    randmove = random()%4;
    return (randmove + 1);
  }
}





void create_enemy(int type, int tile)
{
  int q = 0;
  int found = 0;
  while ((q < 15) && (found == 0))
  {
    if (enemy[q].exist == 0)
    {
      enemy[q].exist = 1;
      enemy[q].type = type;
      enemy[q].location = tile;
      enemy[q].oldlocation = tile;
      enemy[q].frame = 0;
      enemy[q].health = typehealth[type - 1];
      enemy[q].speed = typespeed[type - 1];
      enemy[q].power = typepower[type - 1];
      enemy[q].moving = -1;
      enemy[q].movecount = 0;
      enemy[q].direction = 2;
      enemy[q].shielded = 0;
      enemy[q].shieldcount = 0;
      found = 1;
    }
    else
    {
      q++;
    }
  }
}





int check_move(int tilenum, int direction)
{
  int ok = 0;
  int i = 0;
  int destination = 0;
  if (get_tile(tilenum, direction) != -1)
  {
    destination = get_tile(tilenum, direction);
    ok = 1;
  }
/*  switch(direction)
  {
    case 1:
      if (tile[tilenum].up != -1)
      {
        ok = 1;
        destination = tile[tilenum].up;
      }
      break;
    case 2:
      if (tile[tilenum].down != -1)
      {
        ok = 1;
        destination = tile[tilenum].down;
      }
      break;
    case 3:
      if (tile[tilenum].left != -1)
      {
        ok = 1;
        destination = tile[tilenum].left;
      }
      break;
    case 4:
      if (tile[tilenum].right != -1)
      {
        ok = 1;
        destination = tile[tilenum].right;
      }
      break;
  }*/
  if (ok == 1)
  {
    i = 0;
    while ((i < 15) && (ok == 1))
    {
      if (enemy[i].exist == 1)
      {
        if ((enemy[i].oldlocation == destination) || (enemy[i].location == destination))
        {
          ok = 0;
        }
        else if (enemy[i].moving > 0)
        {
          if (get_tile(enemy[i].oldlocation, enemy[i].moving) == destination)
          {
            ok = 0;
          }
          /*switch(enemy[i].moving)
          {
            case 1:
              if (tile[enemy[i].oldlocation].up == destination)
              {
                ok = 0;
              }
              break;
            case 2:
              if (tile[enemy[i].oldlocation].down == destination)
              {
                ok = 0;
              }
              break;
            case 3:
              if (tile[enemy[i].oldlocation].left == destination)
              {
                ok = 0;
              }
              break;
            case 4:
              if (tile[enemy[i].oldlocation].right == destination)
              {
                ok = 0;
              }
              break;
          }*/
        }
      }
      i++;
    }
/*    if (ok == 1)
    {
      if ((player[0].oldlocation == destination) || (player[0].location == destination))
      {
        ok = 0;
      }
      else if (player[0].moving > 0)
      {
        switch(player[0].moving)
        {
          case 1:
            if (tile[player[0].oldlocation].up == destination)
            {
              ok = 0;
            }
            break;
          case 2:
            if (tile[player[0].oldlocation].down == destination)
            {
              ok = 0;
            }
            break;
          case 3:
            if (tile[player[0].oldlocation].left == destination)
            {
              ok = 0;
            }
            break;
          case 4:
            if (tile[player[0].oldlocation].right == destination)
            {
              ok = 0;
            }
            break;
        }
      }*/
      if (ok == 1)
      {
        return direction;
      }
      else
      {
        return -1;
      }
//    }
  }
  else
  {
    return -1;
  }
}





int invert_move(int direction)
{
  switch(direction)
  {
    case 1:
      return 2;
      break;
    case 2:
      return 1;
      break;
    case 3:
      return 4;
      break;
    case 4:
      return 3;
      break;
  }
}





int rotate_move(int direction)  //rotates the move 90 degrees
{
  switch(direction)
  {
    case 1:
      return 4;
      break;
    case 2:
      return 3;
      break;
    case 3:
      return 1;
      break;
    case 4:
      return 2;
      break;
  }
}





void draw_spells()
{
  int a;
  int i = 0;
  int t;
  int offset_x;
  int offset_y;
  int spell_length;
  int type;
  while (i < 5)
  {
    if (spell[i].active == 1)                      //this convoluted mess draws the
    {                                              //percentage bar and (if % == 100)
      type = spell[i].type - 1;                    //displays the spell name in a box
      if (spell[i].time < archive[type].time)
      {
        spell[i].time += spell[i].speed;
      }
      else if (spell[i].time > archive[type].time)
      {
        spell[i].time = archive[type].time;
      }
      spell[i].percent = ((spell[i].time * 100) / archive[type].time);
      spell_length = (spell[i].percent * 141) / 100;
      rectfill(back_buffer, 479, 8 + (30 * i), 480 + spell_length, 30 + (30 * i), 2);
      if (spell[i].percent == 100)
      {
        int l = text_length(font, spell[i].name);
        rect(back_buffer, 548 - (l / 2), 14 + (30 * i), 550 + (l / 2), 23 + (30 * i), 4);
        rectfill(back_buffer, 549 - (l / 2), 15 + (30 * i), 549 + (l / 2), 22 + (30 * i), 0);
        textout_centre(back_buffer, font, spell[i].name, 550, 15 + (30 * i), 15);
      }
    }
    draw_sprite(back_buffer, (BITMAP *)statbar[0].dat, 475, 5 + (30 * i));
    i++;
  }
  i = 0;
  while (i < 10)
  {
    if (ranged[i].exist == 1)
    {
      ranged[i].movecount += ranged[i].speed;
      if (ranged[i].movecount > 22)
      {
        ranged[i].oldlocation = ranged[i].location;
        ranged[i].movecount -= 22;
      }      
      switch(ranged[i].moving)
      {
        case 1:
          offset_x = -1 * ranged[i].movecount;
          offset_y = -1 * ranged[i].movecount;
          if (ranged[i].movecount > 11)
          {
            ranged[i].location = tile[ranged[i].oldlocation].up;
          }
          break;
        case 2:
          offset_x = ranged[i].movecount;
          offset_y = ranged[i].movecount;
          if (ranged[i].movecount > 11)
          {
            ranged[i].location = tile[ranged[i].oldlocation].down;
          }
          break;
        case 3:
          offset_x = -1 * ranged[i].movecount;
          offset_y = ranged[i].movecount;
          if (ranged[i].movecount > 11)
          {
            ranged[i].location = tile[ranged[i].oldlocation].left;
          }
          break;
        case 4:
          offset_x = ranged[i].movecount;
          offset_y = -1 * ranged[i].movecount;
          if (ranged[i].movecount > 11)
          {
            ranged[i].location = tile[ranged[i].oldlocation].right;
          }
          break;
      }
      ranged[i].frame++;
      draw_ranged_spell(ranged[i], offset_x, offset_y);
      if (ranged[i].location == -1)
      {
        ranged[i].exist = 0;
      }
      a = 0;
      while (a < 15)
      {
        if (enemy[a].exist == 1)
        {
          if ((enemy[a].location == ranged[i].location))
          {
            t = 0;
            while (t < 8)
            {
              init_particle(2, tile[enemy[a].location].x + 25, tile[enemy[a].location].y + 20, 15);
              t++;
            }
            enemy[a] = deal_damage(enemy[a], ranged[i].power, ranged[i].element);
            if (ranged[i].type == 41)  //when glacier hits an enemy
            {                          //the enemy becomes frozen
              if (enemy[a].exist == 1)
              {
                tile[enemy[a].location].status = 23;  //use subzero to trap the enemy
                tile[enemy[a].location].spell = 23;
                enemy[a].trapped = 1;
              }
            }
            ranged[i].exist = 0;
          }
        }
        a++;
      }
    }
    i++;
  }
  i = 0;
  while (i < 81)
  {
    if (graphic[i].exist == 1)
    {
      graphic[i].temp++;
      if (graphic[i].temp > graphic[i].max)
      {
        graphic[i].temp = 0;
      }
      switch(graphic[i].type)
      {
        case 1:  //fire
          draw_sprite(back_buffer, (BITMAP *)fire[graphic[i].temp].dat, graphic[i].x, graphic[i].y);
          break;
        case 11:  //lightning
          draw_sprite(back_buffer, (BITMAP *)lightning[graphic[i].temp].dat, graphic[i].x, graphic[i].y);
          break;
      }
      graphic[i].life--;
      if (graphic[i].life <= 0)
      {
        graphic[i].exist = 0;
        graphic[i].max = 0;
        graphic[i].temp = 0;
      }
    }
    i++;
  }
}





mover cast_enemy_spell(mover current, int type)
{
  spell[5].type = type;
  type = spell[5].type - 1;
  spell[5].speed = archive[type].speed;      //This stuff initializes the spell's
  spell[5].element = archive[type].element;  //attributes.  For the player, these
  spell[5].power = archive[type].power;      //attributes are set when the level
  spell[5].time = archive[type].time;        //begins, but creatures that use spells
  spell[5].name = archive[type].name;        //have to do it this way.
  current = cast_spell(current, 5);
  return current;
}





mover cast_spell(mover current, int spellnum)
{
  int num;
  int tilenum;
  int y = 0;
  int r = 0;
  int temp;
  int temp2;
  spell[spellnum].time = 0;
  switch(current.direction)
  {
    case 1:
      tilenum = tile[current.location].up;
      init_particle(4, tile[current.location].x + 22, tile[current.location].y - 3, 15);
      break;
    case 2:
      tilenum = tile[current.location].down;
      init_particle(4, tile[current.location].x + 20, tile[current.location].y + 4, 15);
      break;
    case 3:
      tilenum = tile[current.location].left;
      init_particle(4, tile[current.location].x + 9, tile[current.location].y + 5, 15);
      break;
    case 4:
      tilenum = tile[current.location].right;
      init_particle(4, tile[current.location].x + 30, tile[current.location].y - 1, 15);
      break;
  }
/*  if (tilenum == -1) //if you are facing blank space and are trying to use something
  {                  //other than a heal spell or a ranged attack that shoots behind
    if ((archive[spell[spellnum].type - 1].kind != 2) && (archive[spell[spellnum].type - 1].kind != 4))
    {
      return current;  //go back without actually doing the spell
    }
  }*/
  int hit = 0;
  int ll;
  int xv;
  int yv;
  int xx;
  int yy;
  switch(spell[spellnum].type)
  {
    case 1:  //fire
    case 4:  //torch
      tile[tilenum].spell = 1;
      r = 0;
      temp = find_free_graphic();
      if (temp != -1)
      {
        graphic[temp].exist = 1;
        graphic[temp].max = 4;
        graphic[temp].life = 25;
        graphic[temp].x = tile[tilenum].x;
        graphic[temp].y = tile[tilenum].y;
        graphic[temp].type = 1;
      }
      r = enemy_on_tile(tilenum);
      if (r != -1)
      {
        y = 0;
        while (y < 60)
        {
          init_particle(2, tile[tilenum].x + 25, tile[tilenum].y + 19, 4);
          y++;
        }
        enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
      }
      break;
    case 15:  //burn
    case 30:  //inferno
      int zmax;
      int zinc;
      int zdir;
      int z[12];  //the z[] holds the locations of the multiple tiles burn hits
      if (spell[spellnum].type == 15)
      {
        zmax = 4;
      }
      else
      {
        zmax = 12;
      }
      zinc = 0;
      zdir = current.direction;
      while (zinc < zmax)
      {
        z[zinc] = get_tile(current.location, zdir);          //  []  <-zdir == 1
        z[zinc + 1] = get_tile(z[zinc], zdir);               //  [][]
        z[zinc + 2] = get_tile(z[zinc], rotate_move(zdir));  //  {}         {}[][]
        zinc += 3;                                           //  zdir == 2->  []
        zdir++;
        if (zdir > 4)
        {
          zdir -= 4;
        }
      }               //If it is just burn, it adds the fourth flame
      if (zmax == 4)  //here->[][][]
      {               //        {}
        z[3] = get_tile(z[0], invert_move(rotate_move(current.direction)));
      }
      temp2 = 0;
      while (temp2 < zmax)
      {
        if (z[temp2] != -1)  //if the spell's extended range doesn't go off the board
        {
          temp = find_free_graphic();
          if (temp != -1)
          {
            graphic[temp].exist = 1;
            graphic[temp].max = 4;
            graphic[temp].life = 25;
            graphic[temp].x = tile[z[temp2]].x;
            graphic[temp].y = tile[z[temp2]].y;
            graphic[temp].type = 1;
          }
          r = enemy_on_tile(z[temp2]);
          if (r != -1)
          {
            y = 0;
            while (y < 40)
            {
              init_particle(2, tile[z[temp2]].x + 25, tile[z[temp2]].y + 19, 4);
              y++;
            }
            enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
          }
        }
        temp2++;
      }
      break;
    case 2:   //spark
    case 16:  //flamejet
    case 21:  //ball-lightning
    case 37:  //fireball
    case 41:  //glacier
      num = find_free_ranged();
      if (num != -1)
      {
        ranged[num].frame = 0;
        ranged[num].exist = 1;
        ranged[num].oldlocation = current.location;
        ranged[num].location = tilenum;
        ranged[num].type = spell[spellnum].type;
        ranged[num].element = spell[spellnum].element;
        if ((spell[spellnum].type == 2) || (spell[spellnum].type == 41))
        {
          ranged[num].speed = 1;
        }
        else if ((spell[spellnum].type == 16) || (spell[spellnum].type == 37))
        {
          ranged[num].speed = 2;
        }
        else
        {
          ranged[num].speed = 3;
        }
        ranged[num].moving = current.direction;
        ranged[num].movecount = 10;
        ranged[num].power = spell[spellnum].power;
      }
      break;
    case 3:  //shock
      temp = 0;
      while (temp < 2)
      {
        num = find_free_ranged();
        if (num != -1)
        {
          ranged[num].exist = 1;
          ranged[num].oldlocation = current.location;
          ranged[num].location = tilenum;
          ranged[num].type = 2;
          ranged[num].speed = 4;
          ranged[num].moving = current.direction;
          ranged[num].element = spell[spellnum].element;
          if (temp == 1)
          {
            ranged[num].moving = invert_move(current.direction);
          }
          ranged[num].movecount = 10;
          ranged[num].power = spell[spellnum].power;
        }
        temp++;
      }
      break;
    case 5:   //jab
    case 10:  //needle
      tile[tilenum].spell = spell[spellnum].type;  //this is there because I use the
      tile[tilenum].status = spell[spellnum].type; //same init code for multiple spells
      r = enemy_on_tile(tilenum);
      if (r != -1)
      {
        y = 0;
        while (y < 60)
        {
          init_particle(2, tile[tilenum].x + 25, tile[tilenum].y + 19, 140);
          y++;
        }
        enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
      }
      break;
    case 6:   //heal
    case 7:   //cure
    case 8:   //replenish
    case 13:  //life
      current.health += spell[spellnum].power;    //increases the caster's health
      if (current.health > MAXHEALTH)             //by the spell's power rating
      {
        current.health = MAXHEALTH;
      }
      init_particle(4, tile[current.location].x + 25, tile[current.location].y + 19, 77);
      if (spell[spellnum].type == 13)  //extra effects for the Life spell
      {
        init_particle(4, tile[current.location].x + 25, tile[current.location].y + 19, 15);
      }
      break;
    case 9:   //spike
    case 12:  //lightning rod
    case 14:  //firepit
    case 18:  //ice
    case 31:  //fireshock
    case 33:  //encase
    case 36:  //amber
      set_trap(tilenum, spell[spellnum].type);
      break;
    case 11:  //lightning
    case 26:  //bolts
      ll = tilenum;
      while (ll != -1)  //this loop checks all the tiles in a line with the player's
      {                 //direction, and also checks whether an enemy is on the tile
        r = enemy_on_tile(ll);
        if (r != -1)
        {
          if (spell[spellnum].type != 26)
          {
            hit = 1;  //the hit variable makes sure only one enemy gets struck
          }
          tile[ll].spell = 11;
          temp = find_free_graphic();
          if (temp != -1)
          {
            graphic[temp].exist = 1;
            graphic[temp].max = 1;
            graphic[temp].life = 20;
            graphic[temp].x = tile[ll].x + 19;
            graphic[temp].y = tile[ll].y - 75;
            graphic[temp].type = 11;
          }
          y = 0;
          while (y < 30)
          {
            init_particle(2, tile[ll].x + 25, tile[ll].y + 19, 15);
            init_particle(2, tile[ll].x + 25, tile[ll].y + 19, 53);  //light blue
            y++;
          }
          enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
        }
        if (hit == 0)
        {
          ll = get_tile(ll, current.direction);
        }
        else
        {
          ll = -1;
        }
      }
      break;
    case 17:  //leech
      r = enemy_on_tile(tilenum);
      if (r != -1)
      {
        ll = invert_move(current.direction);
        switch(ll)
        {
          case 1:
            xv = -1;
            yv = -1;
            break;
          case 2:
            xv = 1;
            yv = 1;
            break;
          case 3:
            xv = -1;
            yv = 1;
            break;
          case 4:
            xv = 1;
            yv = -1;
            break;
        }
        y = 0;
        while (y < 10)
        {
          xx = tile[tilenum].x + 24 + rand()%3;
          yy = tile[tilenum].y + 17 + rand()%3;
          init_particle(7, xx, yy, 46, xv * 3, yv * 3);  //greenish
          y++;
        }
        enemy[r] = deal_damage(enemy[r], 3, 5);
        current.health += 3;
      }
      else  //no enemy
      {
        init_particle(4, tile[tilenum].x + 25, tile[tilenum].y + 25, 46);
      }
      break;
    case 19:  //shield
      current.shielded = 1;
      current.shieldcount = spell[spellnum].power;
      break;
    case 28:  //barrier
      current.shielded = 2;
      current.shieldcount = spell[spellnum].power;
      break;
    case 29:  //power
      current.shielded = 3;
      current.shieldcount = spell[spellnum].power;
      break;
    case 22:  //arctic
    case 23:  //subzero
    case 25:  //plasmic
      int tmax;
      if ((spell[spellnum].type == 22) || (spell[spellnum].type == 25))
      {
        tmax = 5;
      }
      else
      {
        tmax = 9;
      }
      int tset[9];
      int cdir;
      tset[0] = tilenum;                               //      []   <-|
      tset[1] = get_tile(tilenum, current.direction);  //    [][][] <-|--affected tiles
      tset[2] = get_tile(tset[1], current.direction);  //      []   <-|
      cdir = current.direction;                        //      {} <--player
      if (cdir > 2)
      {
        cdir -= 2;  //gets a tile perpendicular to the player's direction
      }
      else
      {
        cdir += 2;  //same
      }
      tset[3] = get_tile(tset[1], cdir);
      tset[4] = get_tile(tset[1], invert_move(cdir));
      if (tmax == 9)                                                 //[][][]
      {                                                              //[][][]
        tset[5] = get_tile(tset[3], current.direction);              //[][][]
        tset[6] = get_tile(tset[3], invert_move(current.direction)); //  {}
        tset[7] = get_tile(tset[4], current.direction);
        tset[8] = get_tile(tset[4], invert_move(current.direction));
      }
      y = 0;
      while (y < tmax)
      {
        if (tset[y] != -1)
        {
          set_trap(tset[y], spell[spellnum].type);  //sets the spell on the tiles
        }
        y++;
      }
      break;
    case 27:  //thunder
    case 35:  //thunderstrike
      y = 0;
      while (y < 81)
      {
        if (spell[spellnum].type == 27)
        {
          hit = random()%2;
        }
        else
        {
          hit = 1;
        }
        if (hit == 1)
        {
          temp = 0;
          r = enemy_on_tile(y);
          if (r != -1)
          {
            enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
            temp = 1;
          }
          else
          {
            r = player_on_tile(y);
            if ((r != -1) && (r != 0))
            {
              player[r] = deal_damage(player[r], spell[spellnum].power, spell[spellnum].element);
              temp = 1;
            }
          }
          if (spell[spellnum].type == 27)
          {
            if (current.location == y)  //it has to check the caster separately
            {                           //because of the way cast_spell works
              current = deal_damage(current, spell[spellnum].power, spell[spellnum].element);
              temp = 1;
            }
          }
          if (temp == 1)
          {
            tile[y].spell = 11;
            temp = find_free_graphic();
            if (temp != -1)
            {
              graphic[temp].exist = 1;
              graphic[temp].max = 1;
              graphic[temp].life = 20;
              graphic[temp].x = tile[y].x + 19;
              graphic[temp].y = tile[y].y - 75;
              graphic[temp].type = 11;
            }
          }
        }
        y++;
      }
      break;
    case 32:  //earthquake
      temp = 1;
      while (temp <= 4)  //this sends the line of quakes in every direction
      {
        ll = get_tile(current.location, temp);
        while (ll != -1)  //this loop checks all the tiles in a line with the player's
        {                 //direction, and also checks whether an enemy is on the tile
          tile[ll].spell = spell[spellnum].type;
          tile[ll].status = spell[spellnum].type;
          y = 0;
          while (y < 15)
          {
            init_particle(2, tile[ll].x + 25, tile[ll].y + 19, 140);
            y++;
          }
          r = enemy_on_tile(ll);
          if (r != -1)
          {
            y = 0;
            while (y < 15)
            {
              init_particle(2, tile[ll].x + 25, tile[ll].y + 19, 140);
              y++;
            }
            enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
          }
          else
          {
            r = player_on_tile(ll);
            if (r != -1)
            {
              y = 0;
              while (y < 15)
              {
                init_particle(2, tile[ll].x + 25, tile[ll].y + 19, 140);
                y++;
              }
              player[r] = deal_damage(player[r], spell[spellnum].power, spell[spellnum].element);
            }
          }
          ll = get_tile(ll, temp);
        }
        temp++;
      }
      break;
    case 34:  //world-tremor
      ll = 0;
      while (ll < 81)
      {
        tile[ll].spell = 34;
        tile[ll].status = 34;
        r = enemy_on_tile(ll);
        if (r != -1)
        {
          enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
        }
        else
        {
          r = player_on_tile(ll);
          if (r != -1)
          {
            player[r] = deal_damage(player[r], spell[spellnum].power, spell[spellnum].element);
          }
        }
        if (current.location == ll)  //it has to check the caster separately
        {                            //because of the way cast_spell works
          current = deal_damage(current, spell[spellnum].power, spell[spellnum].element);
        }
        ll++;
      }
      break;
    case 38:  //geos
      int a[4];
      a[0] = tilenum;
      a[1] = get_tile(a[0], current.direction);
      a[2] = get_tile(a[0], rotate_move(current.direction));
      a[3] = get_tile(a[0], invert_move(rotate_move(current.direction)));
      ll = 0;
      while (ll < 4)
      {
        if (a[ll] != -1)
        {
          y = 0;
          while (y < 15)
          {
            init_particle(9, tile[a[ll]].x + 10 + random()%30, tile[a[ll]].y + 10 + random()%30, 0);
            y++;
          }
          r = enemy_on_tile(a[ll]);
          if (r != -1)
          {
            enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
          }
          tile[a[ll]].spell = 32;  //uses the earthquake graphic on the tile
          tile[a[ll]].status = 32;
        }
        ll++;
      }
      break;
    case 40:  //armageddon
      y = 0;
      while (y < 81)
      {
        hit = random()%2;
        if (hit == 1)
        {
          temp = 0;
          r = enemy_on_tile(y);
          if (r != -1)
          {
            enemy[r] = deal_damage(enemy[r], spell[spellnum].power, spell[spellnum].element);
            temp = 1;
          }
          else
          {
            r = player_on_tile(y);
            if ((r != -1) && (r != 0))
            {
              player[r] = deal_damage(player[r], spell[spellnum].power, spell[spellnum].element);
              temp = 1;
            }
          }
          if (temp == 1)
          {
            temp = find_free_graphic();
            if (temp != -1)
            {
              graphic[temp].exist = 1;
              graphic[temp].max = 4;
              graphic[temp].life = 25;
              graphic[temp].x = tile[y].x;
              graphic[temp].y = tile[y].y;  //use the fire graphic
              graphic[temp].type = 1;
            }
            tile[y].status = 14;  //also use the firepit graphic
            tile[y].spell = 14;
          }
        }
        y++;
      }
      break;
    case 42:  //omni
      init_particle(10, 320, 240, 15);
      y = 0;
      while (y < 15)
      {
        if (enemy[y].exist == 1)
        {
          enemy[y] = deal_damage(enemy[y], spell[spellnum].power, spell[spellnum].element);
        }
        y++;
      }
      break;
  }
  return current;
}





void initialize_spells()
{
  int i = 0;
  int type;
  while (i < 5)
  {
    if (spell[i].active == 1)
    {
      type = spell[i].type - 1;
      spell[i].speed = archive[type].speed;
      spell[i].element = archive[type].element;
      spell[i].power = archive[type].power;
      spell[i].time = archive[type].time;
      spell[i].name = archive[type].name;
    }
    i++;
  }
}





mover deal_damage(mover current, int pow, int element)
{
  int p = 0;
  int s;
  if (current.type != 0)  //if current is not the player
  {
    if (element != 5)  //if it is an elemental attack (which most spells are)
    {
      if (enemydisc[current.type - 1].strong[element - 1] == 1)
      {  //if the enemy is strong against this element
        pow /= 2;  //pow is sliced in half
      }
      if (enemydisc[current.type - 1].weak[element - 1] == 1)
      {  //if the enemy is weak against this element
        pow += (pow / 2);  //pow is upped by half
      }
    }
  }
  if ((current.shielded != 2) && (current.shielded != 3))
  {                         //since deal_damage is only invoked for spell attacks
    current.health -= pow;  //it only deals damage if current doesn't have the
  }                         //barrier or power spells on
  if (current.health <= 0)
  {
    current.exist = 0;
    p = 0;
    while (p < 30)
    {
      if (p < 20)
      {
        init_particle(3, tile[current.location].x + 24 + random()%3, tile[current.location].y + 18 + random()%3, 0);
      }
      init_particle(2, tile[current.location].x + 25, tile[current.location].y + 19, 4);
      p++;
    }
    if (current.type != 0)  //If the player is alive, drops an item where current
    {                       //died.  This way, the player doesn't leave an item
      p = random()%5;       //behind if he gets killed.
      if (p == 1)
      {
        drop_item(current.location, -1);  //drop a random item
      }
      s = numpoints(current.type);  //It also adds to the score if it
      score += s;                   //wasn't the player that was killed
    }
  }
  else
  {
    p = 0;
    while (p < 20)
    {
      init_particle(2, tile[current.location].x + 25, tile[current.location].y + 19, 4);
      init_particle(2, tile[current.location].x + 25, tile[current.location].y + 19, 12);
      p++;
    }
  }
  return current;
}





int find_free_ranged()
{
  int i = 0;
  while (i < 10)
  {
    if (ranged[i].exist == 0)
    {
      return i;
    }
    i++;
  }
  return -1;
}





int find_free_graphic()
{
  int i = 0;
  while (i < 81)
  {
    if (graphic[i].exist == 0)
    {
      return i;
    }
    i++;
  }
  return -1;
}





void draw_ranged_spell(rspell current, int offset_x, int offset_y)
{
  int w;
  int x;
  int y;
  switch(current.type)
  {
    case 2:  //spark
      w = 0;
      x = tile[current.oldlocation].x + offset_x + 25;
      y = tile[current.oldlocation].y + offset_y + 23;
      while (w < 5)
      {
        init_particle(2, x, y, 14);
        w++;
      }
      break;
    case 16:  //flamejet
    case 37:  //fireball
      int px;
      int py;
      switch(current.moving)
      {
        case 1:
          offset_x += 15;
          offset_y += 15;
          px = -1;
          py = -1;
          break;
        case 2:
          offset_x += 10;
          offset_y += 10;
          px = 11;
          py = 11;
          break;
        case 3:
          offset_x += 15;
          offset_y += 20;
          px = 0;
          py = 11;
          break;
        case 4:
          offset_x += 20;
          offset_y += 20;
          px = 12;
          py = -1;
          break;
      }
      x = tile[current.oldlocation].x + offset_x;
      y = tile[current.oldlocation].y + offset_y;
      if (current.type == 16)
      {
        draw_sprite(back_buffer, (BITMAP *)flamejet[current.moving - 1].dat, x, y);
        w = 0;
        while (w < 6)
        {
          init_particle(8, x + px + random()%5, y + py + random()%4, 0);
          w++;
        }
      }
      else
      {
        draw_sprite(back_buffer, (BITMAP *)fireball[0].dat, x, y);
        w = 0;
        while (w < 17)
        {
          init_particle(8, x + random()%15, y + random()%15, 0);
          w++;
        }
      }
      break;
    case 21:  //ball-lightning
      x = tile[current.oldlocation].x + offset_x + 18;
      y = tile[current.oldlocation].y + offset_y + 13;
      draw_sprite(back_buffer, (BITMAP *)ball_lit[current.frame%4].dat, x, y);
      break;
    case 41:  //glacier
      x = tile[current.oldlocation].x + offset_x;
      y = tile[current.oldlocation].y + offset_y;
      draw_sprite(back_buffer, (BITMAP *)ice[0].dat, x, y - 7);
      break;
  }
}





void check_for_win()
{
  static int timer = 0;
  int i = 0;
  int counting = 1;
  while (i < 15)
  {
    if (enemy[i].exist == 1)
    {
      counting = 0;
    }
    i++;
  }
  if (counting == 1)
  {
    timer++;
    if (timer > 100)
    {
      kill_particles();
      gamestatus = 7;
      timer = 0;
    }
  }
}





int check_for_loss()
{
  int w;
  static int timer = 0;
  if (player[0].health <= 0)
  {
    init_particle(3, tile[player[0].location].x + 25, tile[player[0].location].y + 25, 0);
    if (timer == 0)
    {
      init_particle(4, tile[player[0].location].x + 25, tile[player[0].location].y + 15, 44);
    }
    w = 0;
    died = 1;
    timer++;
    if (timer > 100)
    {
      w = check_highscores();
      if (w == -1)
      {
        gamestatus = 0;  //sends you to the menu
      }
      else if (w > -1)
      {
        gamestatus = 5;  //sends you to the high-score post
      }
      timer = 0;
    }
    return 1;
  }
  else
  {
    return 0;
  }
}





int numpoints(int type)
{
  switch(type)
  {
    case 1:
      return 10;
      break;
    case 2:
      return 25;
      break;
  }
}





void query_exit()
{
  int q = 0;
  int w = 0;
  rectfill(screen, 200, 200, 440, 300, 0);
  rect(screen, 200, 200, 440, 300, 2);
  textout_centre(screen, font, "Stop playing?", 320, 210, 15);
  textout_centre(screen, font, "ENTER: Yes, I'm a quitter!", 320, 250, 15);
  textout_centre(screen, font, "ESCAPE: Never!!!", 320, 270, 15);
  while (key[KEY_ESC])
  {
  }
  while (q == 0)
  {
    if (key[KEY_ENTER])
    {
      kill_particles();
      gamestatus = 0;
      keydelay = 5;
      q = 1;
    }
    else if (key[KEY_ESC])
    {
      keydelay = 5;   //a special thingy to keep the escape key from being read
      q = 1;          //immediately after the program returns to the main game loop
    }
  }
  counter = 0;
}





void load_game()
{
  static int choice = 0;
  static int delay = 0;
  static string savedname[10];
  int q;
  int done;
  clear(back_buffer);
  if (initialized == 0)
  {
    q = 0;
    while (q < 10)
    {
      sprintf(msg, "player%-1d.sav", q);
      if (exists(msg))
      {
        ifstream savedgame(msg);        //if the file exists, load its name
        savedgame >> savedname[q];      //property onto the savedname array
      }
      else
      {
        savedname[q] = "-EMPTY-";       //if it doesn't exist, make it "-EMPTY-"
      }
      q++;
    }   
    initialized = 1;
  }
  textout_centre(back_buffer, (FONT *)fonts[0].dat, "Load Game", 320, 50, 15);
  textout(back_buffer, font, "Press ESC to return to the title screen", 10, 460, 2);
  q = 0;
  static int nkeydelay = 10;
  int loc_x = 220;           //all this stuff is for choosing
  int loc_y = 150;           //the save file to load from
  int lx = loc_x;
  int ly = loc_y;
  rect(back_buffer, loc_x - 5, loc_y - 10, loc_x + 205, loc_y + 75, 15);
  rectfill(back_buffer, loc_x + 1, loc_y - 9, loc_x + 204, loc_y + 74, 0);
  ly = 145 + (15 * choice);                          //this
  if (choice > 4)                                    //code
  {                                                  //draws
    lx += 100;                                       //the
    ly -= 75;                                        //file
  }                                                  //selection
  rect(back_buffer, lx, ly, lx + 100, ly + 15, 14);  //rectangle
  while (q < 10)
  {
    textout_centre(back_buffer, font, savedname[q].c_str(), loc_x + 50, loc_y, 15);
    loc_y += 15;
    if (loc_y >= 225)
    {
      loc_y = 150;
      loc_x += 100;
    }
    q++;
  }
  if (nkeydelay == 0)
  {
    if ((key[KEY_UP]) && (choice > 0))
    {
      choice--;
      nkeydelay = 10;
    }
    if ((key[KEY_DOWN]) && (choice < 9))
    {
      choice++;
      nkeydelay = 10;
    }
    if ((key[KEY_RIGHT]) && (choice < 5))
    {
      choice += 5;
      nkeydelay = 10;
    }
    if ((key[KEY_LEFT]) && (choice > 4))
    {
      choice -= 5;
      nkeydelay = 10;
    }
    if (key[KEY_ENTER])
    {
      if (savedname[choice] != "-EMPTY-")
      {
        sprintf(msg, "player%-1d.sav", choice); //actually load the info
        ifstream loader(msg);
        loader >> savedname[choice];
        q = 0;
        while (q < TOTAL_SPELLS)
        {
          loader >> archive[q].active;
          q++;
        }
        loader >> level;
        q = 0;
        while (q < 4)
        {
          loader >> orb[q];
          q++;
        }
        loader >> player[0].health;
        loader >> warp;
        q = 0;
        while (q < 5)
        {
          loader >> spell[q].type;
          q++;
        }
        loader >> score;
        loader >> MAXHEALTH;
        loader >> nextlevel;
        loader >> warplevel;
        loader >> warping;
        q = 0;
        while (q < 5)
        {
          loader >> award[q].oldpoints;
          loader >> award[q].old;
          q++;
        }
        gamestatus = 1;
        keydelay = 10;
        nkeydelay = 10;
        choice = 0;
      }
    }
    if (key[KEY_ESC])
    {
      gamestatus = 0;
      choice = 0;
      keydelay = 10;
      nkeydelay = 10;
    }
  }
  else
  {
    nkeydelay--;
  }
}





void kill_particles()
{
  int z = 0;
  while (z < PARTICLENUM)
  {
    particle[z].exist = 0;
    z++;
  }
}





void show_chosenspell(int spl)
{
  int i;
  int color;
                                                //xmin = 236
                                                //ymin = 41
                                                //xmax = 619
                                                //ymax = 199
  draw_sprite(back_buffer, (BITMAP *)menupic[spl].dat, 250, 50);
  textout(back_buffer, (FONT *)fonts[1].dat, archive[spl].name, 350, 45, 15);
  i = 0;
  while (i < 5)
  {
    if (archive[spl].description[i] != "-")
    {
      textout(back_buffer, font, archive[spl].description[i].c_str(), 240, 145 + (10 * i), 15);
    }
    i++;
  }
  if (archive[spl].active == 0)
  {
    i = 0;
    while (i < 4)
    {
      draw_sprite(back_buffer, (BITMAP *)orbs[i].dat, 240 + (50 * i), 195);
      if (check_price(spl) == 1)
      {
        color = 15;  //white if spell is buyable
      }
      else
      {
        color = 12;  //red if spell requires more orbs than the player has
      }
      sprintf(msg, "%-1d", archive[spl].buy[i]);
      textout(back_buffer, font, msg, 260 + (50 * i), 200, color);
      i++;
    }
  }
}





int increment_chosenspell(int spl, int inc, int buy)
{
  int step = 1;
  int total = spl;
  int w = 0;
  int i;
  int ok;
  while (w == 0)
  {
    total += inc;
    if (total >= TOTAL_SPELLS)
    {
      total = 0;
    }
    else if (total < 0)
    {
      total = TOTAL_SPELLS - 1;
    }
    if (archive[total].active == buy)  //if buy is 1, returns bought spells
    {                                  //if buy is 0, returns buyable spells
      if (buy == 0)  //if the player is looking for buyable spells
      {
        i = 0;
        ok = 1;
        while ((i < TOTAL_SPELLS) && (ok == 1))  //checks to see if the player has
        {                                        //learned the prerequisite spells
          if (archive[total].need[i] != -1)  //if there are more needed spells
          {
            if (archive[archive[total].need[i] - 1].active == 0)
            {
              ok = 0;  //if the prerequisite has not been learned yet
            }
          }
          else
          {
            i = TOTAL_SPELLS;
          }
          i++;
        }
        if (ok == 1)
        {
          return total;
        }
      }
      else  //if the player is choosing already bought spells
      {
        if (archive[total].kind != 7)  //if it isn't a menu-only spell
        {
          return total;
        }
      }
    }
    else
    {
      step++;
      if ((buy == 0) && (step >= TOTAL_SPELLS))
      {
        return spl;  //returns the original spell found if no others are there
      }
    }
    if (key[KEY_B])
    {
      w = 1;  //insurance against an infinite loop
    }
  }
}





int find_first_learner()
{
  int t = 0;
  int i;
  int ok;
  while (t < TOTAL_SPELLS)
  {
    if (archive[t].active == 0)
    {
      i = 0;
      ok = 1;
      while ((i < TOTAL_SPELLS) && (ok == 1))
      {
        if (archive[t].need[i] != -1)  //if there are more required spells
        {
          if (archive[archive[t].need[i] - 1].active == 0)  //if the prerequisite has
          {                                                 //not been learned yet
            ok = 0;
          }
        }
        else
        {
          i = TOTAL_SPELLS;
        }
        i++;
      }
      if (ok == 1)
      {
        return t;
      }
    }
    else
    {
      t++;
    }
  }
  return -1;
}





int check_price(int spl)
{
  int i = 0;
  int ok = 1;
  while (i < 4)
  {
    if (archive[spl].buy[i] > orb[i])
    {
      ok = 0;
    }
    i++;
  }
  if (archive[spl].active == 1)  //if this weren't here, buying somthing and holding
  {                              //down enter would drain your orb supply to nothing
    ok = 0;
  }
  return ok;
}





void buy_spell(int spl)
{
  int i = 0;
  while (i < 4)
  {
    orb[i] -= archive[spl].buy[i];
    i++;
  }
  archive[spl].active = 1;
}





void drop_item(int loc, int type)  //a -1 type indicates random
{
  if (type == -1)
  {
    type = rand()%TOTAL_ITEMS;
  }
  tile[loc].object = type;
}





void grab_item(int type)
{
  switch(type)
  {
    case 0:
      orb[0]++;
      break;
    case 1:
      orb[1]++;
      break;
    case 2:
      orb[2]++;
      break;
    case 3:
      orb[3]++;
      break;
    case 4:
      orb[0]++;
      orb[1]++;
      orb[2]++;
      orb[3]++;
      break;
    case 5:
      score += 500;
      break;
    case 6:
      score += 1000;
      break;
  }
}





void end_of_level()
{
  clear(back_buffer);
  static int time = 0;
  static int first = 0;
  static int c = 32;
  static int d = 1;
  int i;
  if (first == 0)
  {
    i = 0;
    while (i < 5)
    {
      award[i] = get_awards(award[i], i);
      i++;
    }
    if (player[0].health == MAXHEALTH)
    {
      add_warp();
      if (warping == 0)  //if the player has not just finished a warp level
      {
        warping = check_warp(1);  //see the function to see why the '1' is there
      }
    }    
    first = 1;
  }
  textout_centre(back_buffer, (FONT *)fonts[1].dat, "Bonus Points", 320, 50, 15);
  i = 0;
  while (i < time / 30)
  {
    textout(back_buffer, font, award[i].description.c_str(), 220, 100 + (i * 20), 15);
    sprintf(msg, "%-1d", award[i].points);
    textout(back_buffer, font, msg, 400, 100 + (i * 20), 15);
    i++;
  }
  if (time == 150)
  {
    if (player[0].health == MAXHEALTH)
    {
      textout_centre(back_buffer, font, "Full health!", 320, 240, 15);
    }
    draw_sprite(back_buffer, (BITMAP *)misc[0].dat, 220, 250);  //draw the warp pic
    switch(warp)  //no alternative to using this hideously hackish code
    {
      case 0:
        rectfill(back_buffer, 220, 250, 450, 300, 0);
        break;
      case 1:
        rectfill(back_buffer, 280, 250, 450, 300, 0);
        break;
      case 2:
        rectfill(back_buffer, 330, 250, 450, 300, 0);
        break;
      case 3:
        rectfill(back_buffer, 370, 250, 450, 300, 0);
        break;
    }
    if (warping == 1)
    {
      c += d;
      if ((c > 60) || (c < 33))  //top color is 30, bottom is 17
      {
        d *= -1;
      }
      rect(back_buffer, 210, 255, 430, 300, c / 2);  // the / 2 slows it down
    }
  }
  if (time < 150)
  {
    time++;
  }
  else if (key[KEY_ENTER])
  {
    i = 0;
    while (i < 5)
    {
      score += award[i].points;
      i++;
    }
    first = 0;
    time = 0;
    level++;
    if (level%15 == 0)
    {
      MAXHEALTH += 5;  //up the maximum health the player can have
    }
    if (level == LAST_LEVEL)
    {
      levelinit();  //a bit of a roundabout way, but this checks the highscores
    }               //and sends you to the high-score post or the title screen
    else
    {
      if (warping == 1)
      {
        warp = 0;  //resets the warps
        nextlevel = level;  //sets the level to return to after the warp
        level = warplevel;
        warping = -1;
      }
      else if (warping == -1)
      {
        level = nextlevel;
        warplevel++;
        warping = 0;
      }
      gamestatus = 1;  //if it isn't the last level, go to the menu
    }
    keydelay = 15;
  }
  else
  {
    textout_centre(back_buffer, font, "PRESS ENTER", 320, 320, 15);
  }
}





awardsgiven get_awards(awardsgiven a, int num)
{
  int i;
  int type;
  int ok;
  int p_add;
  int temp;
  switch(num)
  {
    case 0:  //award 1:  More health left, more points gained, multiplied by level
      a.description = "Health bonus:";
      i = level / 5;
      if (i == 0)
      {
        i = 1;
      }
      a.points = player[0].health * i;
      break;
    case 1:
      a.description = "Level bonus:";
      a.points = level * 15;
      break;
    case 2:  //award 3:  If only a certain type of spell is used (ie: only fire, etc.)
      ok = 1;
      type = spell[0].element;
      while (type == 5)
      {
        type = spell[i].element;
        i++;
      }
      i = 1;
      while (i < 5)
      {
        if (spell[i].active == 1)
        {
          if ((type != spell[i].element) && (spell[i].element != 5))
          {
            ok = 0;
          }
        }
        i++;
      }
      if (ok == 1)
      {
        switch(type)
        {
          case 1:
            a.description = "Pyromancer:";
            p_add = 150;
            break;
          case 2:
            a.description = "Geomancer:";
            p_add = 200;
            break;
          case 3:
            a.description = "Aquamancer:";
            p_add = 400;
            break;
          case 4:
            a.description = "Aeromancer:";
            p_add = 100;
            break;
        }
        if ((type != a.old) || (a.old == -1))
        {                                   //if this same element was the only one
          a.oldpoints = 0;                  //used in a previous round also, the
        }                                   //points given increase each time by p_add
        a.old = type;
        a.points = a.oldpoints += p_add;
      }
      else
      {
        a.description = "-";
        a.oldpoints = 0;
        a.points = 0;
        a.old = -1;
      }
      break;
    case 3:  //gives a bonus if all spells are ranged, direct, trap, etc.
      ok = 1;
      temp = archive[spell[0].type - 1].kind;  //confusing
      i = 1;
      while (temp == 6)
      {
        temp = archive[spell[i].type - 1].kind;
        i++;
      }
      while (i < 5)
      {
        if (spell[i].active == 1)
        {
          type = archive[spell[i].type - 1].kind;  //very confusing
          if ((temp != type) && (type != 6))  //if the spell is not the same type and
          {                                   //is also not a nontype spell, eg: shield
            ok = 0;
          }
        }
        i++;
      }
      if (ok == 1)
      {
        if (temp != a.old)
        {
          a.oldpoints = 0;
        }
        switch (temp)
        {
          case 1:
            a.description = "Direct attacker:";
            p_add = 150;
            break;
          case 2:
            a.description = "Distance attacker:";
            p_add = 100;
            break;
          case 3:
            a.description = "Trapper:";
            p_add = 200;
            break;
          case 4:
            a.description = "All heal spells, HOW DID YOU WIN?!?";  //never will happen
            p_add = 1000000;
            break;
          case 5:
            a.description = "Summoner:";
            p_add = 250;
            break;
        }
        a.points = a.oldpoints += p_add;
        a.old = temp;
      }
      else
      {
        a.description = "-";
        a.points = 0;
        a.oldpoints = 0;
        a.old = -1;
      }
      break;
    case 4:  //awards a bonus for the time taken
      gametime /= 50;  //divides the game time by the fps to get seconds played
      if (gametime < 12)  //one level gone in under 12 seconds
      {
        a.description = "Speed Demon:";
        a.points = 500;
      }
      else if (gametime > 240)  //the level took over 2 minutes
      {
        a.description = "Way Too Slow:";
        a.points = 1;
      }
      else
      {
        a.description = "Time Bonus:";
        a.points = 240 - gametime;
      }
      break;
  }
  return a;
}





void add_warp()
{
  int w = check_warp(0);
  if (w < 4)  //it always should be less than 4, but this makes it certain
  {
    warp++;
  }
}





int check_warp(int func)
{
  int i = 0;
  int ok;
  if (warp == 4)
  {
    ok = 1;
  }
  else
  {
    ok = 0;
  }
  if (func == 0)
  {
    return warp;
  }
  else
  {
    return ok;
  }
}





int get_tile(int tilenum, int direction)
{
  if (tilenum != -1)
  {
    switch(direction)
    {
      case 1:
        return tile[tilenum].up;
        break;
      case 2:
        return tile[tilenum].down;
        break;
      case 3:
        return tile[tilenum].left;
        break;
      case 4:
        return tile[tilenum].right;
        break;
    }
  }
  else
  {
    return -1;
  }
}





int enemy_on_tile(int tilenum)
{
  int r = 0;
  while (r < 15)
  {
    if ((enemy[r].exist == 1) && (enemy[r].location == tilenum))
    {
      return r;
    }
    r++;
  }
  return -1;
}





int player_on_tile(int tilenum)
{
  int r = 0;
  while (r < 40)
  {
    if ((player[r].exist == 1) && (player[r].location == tilenum))
    {
      return r;
    }
    r++;
  }
  return -1;
}





void do_description()
{
  static int delay = 0;
  static int selected = 0;
  string thing;
  char name[15];
  int i;
  int c;
  rect(back_buffer, 200, 150, 460, 350, 43);
  rectfill(back_buffer, 201, 151, 459, 349, 0);
  if (delay > 0)
  {
    delay--;
  }
  else
  {
    if (key[KEY_RIGHT])
    {
      selected++;
      if (selected > (NUM_ENEMIES - 1))
      {
        selected = 0;
      }
      delay = 10;
    }
    if (key[KEY_LEFT])
    {
      selected--;
      if (selected < 0)
      {
        selected = NUM_ENEMIES - 1;
      }
      delay = 10;
    }
  }
  draw_sample_enemy(selected + 1, 210, 160);
  switch(selected)
  {
    case 0:  //slime
      sprintf(name, "Slime");
      break;
    case 1:  //goblin
      sprintf(name, "Goblin");
      break;
    case 2:  //mole
      sprintf(name, "Mole");
      break;
    case 3:
      sprintf(name, "Clayman");
      break;
    case 4:
      sprintf(name, "Phantom");
      break;
    case 5:
      sprintf(name, "Imp");
      break;
  }
  textout(back_buffer, (FONT *)fonts[1].dat, name, 270, 160, 15);
  i = 0;
  while (i < 4)
  {
    thing = enemydisc[selected].disc[i];
    if (thing != "-")
    {
      textout(back_buffer, font, thing.c_str(), 210, 210 + (10 * i), 15);
    }
    i++;
  }
  textout(back_buffer, font, "Strong Against:", 210, 260, 15);
  i = 0;
  while (i < 4)
  {
    if (enemydisc[selected].strong[i] == 1)
    {
      c = 15;
    }
    else
    {
      c = 23;  //dark grey
    }
    thing = element(i + 1);
    textout(back_buffer, font, thing.c_str(), 335, 260 + (10 * i), c);
    i++;
  }
  textout(back_buffer, font, "Weak Against:", 210, 300, 15);
  i = 0;
  while (i < 4)
  {
    if (enemydisc[selected].weak[i] == 1)
    {
      c = 15;
    }
    else
    {
      c = 23;  //dark grey
    }
    thing = element(i + 1);
    textout(back_buffer, font, thing.c_str(), 335, 300 + (10 * i), c);
    i++;
  }  
}





string element(int type)
{
  switch(type)
  {
    case 1:
      return "Fire";
      break;
    case 2:
      return "Earth";
      break;
    case 3:
      return "Water";
      break;
    case 4:
      return "Wind";
      break;
  }
}





void set_trap(int tilenum, int type)
{
  tile[tilenum].spell = type;  //gives the trap spell to activate
  tile[tilenum].status = -2;  //makes the tile act as a trap
  int r = 0;
  while (r < 50)
  {
    init_particle(5, tile[tilenum].x + 25, tile[tilenum].y + 19, 39);
    r++;
  }
}





void draw_sample_enemy(int type, int x, int y)
{
  switch(type)
  {
    case 1:
      draw_sprite(back_buffer, (BITMAP *)slime[0].dat, x, y);
      break;
    case 2:
      draw_sprite(back_buffer, (BITMAP *)gobwalk2[2].dat, x, y);
      break;
    case 3:
      draw_sprite(back_buffer, (BITMAP *)mole[0].dat, x, y);
      break;
    case 4:
      draw_sprite_h_flip(back_buffer, (BITMAP *)claydl[0].dat, x, y);
      break;
    case 5:
      draw_sprite(back_buffer, (BITMAP *)phantom[2].dat, x, y);
      break;
    case 6:
      draw_sprite(back_buffer, (BITMAP *)imp2[0].dat, x, y);
      break;
  }
}
